/*******************************************************************
 *
 *	Copyright (c) 1994-1999 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	File: kblock.cc
 *	Revision:  $Id: kblock.cpp,v 1.1 2005/05/12 09:51:39 crypt Rel-1.6-3 $
 *	Created:
 *	Description: implementation of procedures to work with 
 *		Data Block that stores encryption keys
 *
 *******************************************************************/

#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <bc_types.h>
#include "kg_defs.h"
#include "kg_misc.h"
#include "kblock.h"
#include "random.h"
#include "sha256.h"
#include "alg.h"

char kblock_cpp[]="$Id: kblock.cpp,v 1.1 2005/05/12 09:51:39 crypt Rel-1.6-3 $";


/* Forward declarations of local procedures */

BOOL  CalculateDigest( BYTE *Data, DWORD Length, BYTE *Digest );
BOOL  DataBlockVerifyDigest( DATA_BLOCK *DataBlock );
BOOL  DataBlockUpdateDigest( DATA_BLOCK *DataBlock );
BOOL  GetStorageKeyFromPassword( char *Password, BYTE *StorageKey );
void  setAttribute( DWORD *AttBuffer, DWORD Attribute );
DWORD getAttribute( DWORD AttBuffer );
BOOL createAlternativeKeyBlock(
                              DATA_BLOCK *DataBlock,
                              ALG_SERV   Alg,
                              DWORD      AlgKeyLength,
                              BYTE       *Key,
                              BYTE       *PoolBuffer
                              );

/**********************************************************************
 * 
 * DataBlockAllocate() - allocates memory for DATA_BLOCK structure 
 *                       and initializes the structure.
 *
 * The procedure which calls DataBlockAllocate() is responsible for 
 * releasing the memory with the DataBlockFree() function.
 *
 **********************************************************************/

extern BOOL DataBlockAllocate(DATA_BLOCK **DataBlock, 
                              DWORD *DataSize, 
                              DWORD AlgId)
{
    DATA_BLOCK  *db;
    KEY_BLOCK   *kb;
    int         i;

    *DataSize = 0;
    *DataBlock = (DATA_BLOCK *)malloc(sizeof(DATA_BLOCK));

    if ( !(*DataBlock) ) return FALSE;

    db = *DataBlock;
    *DataSize = sizeof(DATA_BLOCK);

    ShredData( (BYTE *)db, sizeof(DATA_BLOCK) ); /* fill DataBlock by random garbage */

    strcpy( db->header.signature, SIGNATURE03 );
//    strcpy( db->header.signature, SIGNATURE04 );
    db->header.formatVersion = FORMAT_VERSION;
    db->header.encrAlgorithmId = AlgId;
    db->header.hashAlgorithmId = SHA256_ID;
    db->header.dataBlockSize = sizeof(DATA_BLOCK);
    db->header.keyBlockSize = sizeof(KEY_BLOCK);
    db->header.keyBlockNumber = MAXIMUM_NUMBER_OF_KEYS;
    setAttribute( &(db->header.state), STATE_POOL_NOT_INITIALIZED);

    kb = db->keys;

    for ( i=0; i<MAXIMUM_NUMBER_OF_KEYS; i++, kb++ )
    {
        setAttribute( &(kb->keyAttribute), KATTRIBUTE_KEY_EMPTY );
    }

    if ( DataBlockUpdateDigest( *DataBlock ) )
        return TRUE;

    DataBlockFree( DataBlock );
    *DataSize = 0;
    return FALSE;
}


/**********************************************************************
 * 
 * DataBlockFree() - wipes and frees DATA_BLOCK structure 
 *
 * The procedure which calls DataBlockAllocate() is responsible for 
 * releasing the memory with the DataBlockFree() function.
 *
 **********************************************************************/

extern BOOL DataBlockFree( DATA_BLOCK **DataBlock )
{
    ShredData( (BYTE *)*DataBlock, sizeof(DATA_BLOCK) );
    if ( *DataBlock ) free( *DataBlock);
    *DataBlock = NULL;
    return TRUE;
}


/**********************************************************************
 * 
 * DataBlockCheck() - verifies that DATA_BLOCK structure is not damaged
 *                    and checks that DATA_BLOCK stores at least 
 *                    KeysCount keys.
 *
 **********************************************************************/

extern DWORD DataBlockCheck(
                           DATA_BLOCK *DataBlock,
                           DWORD      DataSize,
                           DWORD      AlgId
                           )
{
    if ( !DataBlock ) return ERROR_DAMAGED_DATA_BLOCK;

    if ( DataBlock->header.formatVersion != FORMAT_VERSION )
        return ERROR_INCORRECT_FORMAT_VERSION;

    if ( (DataBlock->header.dataBlockSize != sizeof(DATA_BLOCK))
         || (DataSize != sizeof(DATA_BLOCK)) )
        return ERROR_INCORRECT_DATA_BLOCK_SIZE;

    if (strcmp(DataBlock->header.signature, SIGNATURE98) && 
        strcmp(DataBlock->header.signature, SIGNATURE03) &&
        strcmp(DataBlock->header.signature, SIGNATURE04))
        return ERROR_INCORRECT_BLOCK_SIGNATURE;

    if ( DataBlock->header.encrAlgorithmId != AlgId )
        return ERROR_INCORRECT_ENCR_ALGORITHM;

    if ( DataBlock->header.hashAlgorithmId != SHA256_ID )
        return ERROR_INCORRECT_HASH_ALGORITHM;

    if ( !DataBlockVerifyDigest( DataBlock ) )
        return ERROR_DAMAGED_DATA_BLOCK;

    return ERROR_NO;
}



/**********************************************************************
 * 
 * DataBlockWriteKey() - fills a free KEY_BLOCK in the DataBlock.
 *
 **********************************************************************/

extern DWORD DataBlockWriteKey(
                              DATA_BLOCK *DataBlock,
                              ALG_SERV   Alg,
                              DWORD      AlgKeyLength,
                              char *    Password,
                              BYTE       *Key,
                              DWORD      KeyAttribute,
                              BYTE       *PoolBuffer
                              )
{
    BYTE      storageKey[ MAXIMUM_KEY_SIZE_BYTES ];
    DWORD     keyHandle;
    KEY_BLOCK *kb, *kbFound;
    BYTE iVector[ IVECTOR_LENGTH ];

    /* find empty KEY_BLOCK inside DataBlock*/
    kb = DataBlock->keys;
    kbFound = NULL;

    for ( int i=0; i<MAXIMUM_NUMBER_OF_KEYS; i++, kb++ )
    {
        if ( getAttribute(kb->keyAttribute) == KATTRIBUTE_KEY_EMPTY )
        {
            kbFound = kb;
            break;
        }
    }

    if ( !kbFound ) return ERROR_NOT_ENOUGH_SPACE_FOR_KEY;

    /* copy encryption key to the KEY_BLOCK memory */
    memcpy( kbFound->key, Key, MAXIMUM_KEY_SIZE_BYTES );

    /* calculate digest of the key and save it in KEY_BLOCK */
    if ( !CalculateDigest( kbFound->key, MAXIMUM_KEY_SIZE_BYTES, 
                           kbFound->digest ) )
    {
        ShredData( kbFound->key, MAXIMUM_KEY_SIZE_BYTES );
        return ERROR_INVALID_ALGORITHM;
    }

    memcpy( iVector, kbFound->digest, IVECTOR_LENGTH );

    /* Encrypt pool in the data block with the Key */
    if ( getAttribute(DataBlock->header.state) == STATE_POOL_NOT_INITIALIZED )
    {
        /*  we must fill the PoolBuffer with zeros, because 
            the pool in the DataBlock we are going to encrypt 
            already contain random data */
        memset( PoolBuffer, 0, POOL_SIZE_BYTES );
        /* Load the Key to encryption algorithm module */
        if ( alg_make_key (Alg, Key, AlgKeyLength, PoolBuffer, &keyHandle ) )
        {
            ShredData( iVector, IVECTOR_LENGTH);
            return ERROR_INVALID_ALGORITHM;
        }

        /* Encrypt pool */
        if ( alg_encrypt( Alg, keyHandle, iVector, 
                          DataBlock->pool, POOL_SIZE_BYTES ) )
        {
            alg_free_key( Alg, keyHandle );
            ShredData( iVector, IVECTOR_LENGTH);
            return ERROR_INVALID_ALGORITHM;
        }

        /* Erase and free the Key data from the encryption algorithm module */
        alg_free_key( Alg, keyHandle );
        setAttribute(&(DataBlock->header.state),STATE_POOL_INITIALIZED);
    }

    /*	We begin to encrypt file-container's encryption key 
        and its digest using storageKey.
        First of all, calculate storageKey from the password
     */
    if ( !GetStorageKeyFromPassword( Password, storageKey ) )
    {
        ShredData( kbFound->key, MAXIMUM_KEY_SIZE_BYTES );
        ShredData( kbFound->digest, MAXIMUM_DIGEST_SIZE_BYTES );
        ShredData( iVector, IVECTOR_LENGTH);
        return ERROR_INVALID_ALGORITHM;
    }

    /*	Load storageKey into the Algorithm driver and again, 
        the Pool must be filled by zeros, because we are going 
        to encrypt random data (Key) */
    if ( alg_make_key (Alg, storageKey, AlgKeyLength, PoolBuffer, &keyHandle ) )
    {
        ShredData( storageKey, MAXIMUM_KEY_SIZE_BYTES );
        ShredData( kbFound->key, MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES);
        ShredData( iVector, IVECTOR_LENGTH);
        return ERROR_INVALID_ALGORITHM;
    }

    /* Now we don't need in storageKey, because it is copied to the
       Algorithm Driver */
    ShredData( storageKey, MAXIMUM_KEY_SIZE_BYTES );

    /* encrypt the key and its digest copied to the KEY_BLOCK */
    if ( alg_encrypt( Alg, keyHandle, iVector, 
                      kbFound->key, MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES) )
    {
        alg_free_key( Alg, keyHandle );
        ShredData( kbFound->key, MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES);
        ShredData( iVector, IVECTOR_LENGTH);
        return ERROR_INVALID_ALGORITHM;
    }

    /* erase storageKey from the Algorithm driver memory */
    alg_free_key( Alg, keyHandle );

    ShredData( iVector, IVECTOR_LENGTH);

    setAttribute( &(kbFound->keyAttribute), KeyAttribute );

    /* Check if Alternative Key Block is initialized. If no, initialize it */
    createAlternativeKeyBlock(  DataBlock, Alg, AlgKeyLength,Key, PoolBuffer );

    if ( !DataBlockUpdateDigest( DataBlock ) )
    {
        setAttribute( &(kbFound->keyAttribute), KATTRIBUTE_KEY_EMPTY );
        return ERROR_INTERNAL_PROBLEM;
    }

    return ERROR_NO;
}



DWORD DataBlockVerifyPasswordAndGetKey(
                                      DATA_BLOCK *DataBlock,
                                      ALG_SERV Alg,
                                      DWORD      AlgKeyLength,
                                      char *    Password,
                                      BYTE       *Key,
                                      BYTE       *PoolBuffer
                                      )
{
    KEY_BLOCK *kb, *kbFound;
    BYTE  storageKey[ MAXIMUM_KEY_SIZE_BYTES ];
    BYTE  keyAndDigest[ MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES ];
    BYTE  calculatedDigest[ MAXIMUM_DIGEST_SIZE_BYTES ];
    DWORD keyHandle, i;
    BYTE  iVector[ IVECTOR_LENGTH ];

    kb = DataBlock->keys;
    kbFound = NULL;

    /*  we fill the PoolBuffer with zeros, because 
        the buffers in the DataBlock we are going to decrypt 
        contain random data */

    /*	in the first part of the procedure we use the PoolBuffer
        to initialize encryption algorithm module with zeroed pools.
        In the second part we will place a correct decrypted pool 
        data to the PoolBuffer */

    memset( PoolBuffer, 0, POOL_SIZE_BYTES );

    for ( i=0; i<MAXIMUM_NUMBER_OF_KEYS; i++, kb++ )
    {
        if ( getAttribute(kb->keyAttribute) == KATTRIBUTE_KEY_EMPTY ) continue;

        /* Calculate encryption key for decrypting KEY_BLOCK */
        if ( !GetStorageKeyFromPassword( Password, storageKey ) )
            return ERROR_INVALID_ALGORITHM;

        /* Decrypt KEY_BLOCK kb */
        /* First, load storageKey into the Algorithm driver */
        if ( alg_make_key (Alg, storageKey, AlgKeyLength, PoolBuffer, &keyHandle ) )
            return ERROR_INVALID_ALGORITHM;

        /* Now we don't need in storageKey, because it is copied to the
           Algorithm Driver */
        ShredData( storageKey, MAXIMUM_KEY_SIZE_BYTES );

        /* decrypt the key and its digest copied to the KEY_BLOCK */
        memcpy( keyAndDigest, kb->key, MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES );

        /* initialize iVector with any value.
           After first decrypting we will define it exactly */
        memset( iVector, 0, IVECTOR_LENGTH );

        /* first decryption */
        if ( alg_decrypt( Alg, keyHandle, iVector, 
                          keyAndDigest, MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES) )
        {
            alg_free_key( Alg, keyHandle );
            ShredData( keyAndDigest, MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES);
            return ERROR_INVALID_ALGORITHM;
        }

        /* define Initial Vector */
        memcpy( iVector, &(keyAndDigest[MAXIMUM_KEY_SIZE_BYTES]), IVECTOR_LENGTH );

        /* second decryption with exact Initial Vector */
        memcpy( keyAndDigest, kb->key, MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES );

        if ( alg_decrypt( Alg, keyHandle, iVector, 
                          keyAndDigest, MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES) )
        {
            alg_free_key( Alg, keyHandle );
            ShredData( keyAndDigest, MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES);
            ShredData( iVector, IVECTOR_LENGTH);
            return ERROR_INVALID_ALGORITHM;
        }

        /* erase storageKey from the Algorithm driver memory */
        alg_free_key( Alg, keyHandle );

        /* calculate digest of the key and compare it with digest */
        if ( !CalculateDigest( keyAndDigest, MAXIMUM_KEY_SIZE_BYTES, 
                               calculatedDigest ) )
        {
            ShredData( keyAndDigest, MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES);
            ShredData( iVector, IVECTOR_LENGTH);
            return ERROR_INVALID_ALGORITHM;
        }
//-----------------------------------------------------------		
        if ( memcmp( calculatedDigest, 
                     &(keyAndDigest[MAXIMUM_KEY_SIZE_BYTES]), 
                     SHA256_DIGEST_SIZE) == 0 )
        {
            kbFound = kb;
            break;
        }
//    kbFound = kb;
    }

    if ( !kbFound )
    {
        ShredData( iVector, IVECTOR_LENGTH);
        return ERROR_INCORRECT_PASSWORD;
    }

    memcpy( Key, keyAndDigest, MAXIMUM_KEY_SIZE_BYTES );


    /*	In the following part we will decrypt and place a correct 
        pool data to the PoolBuffer,
        Now everything is Ok, and we have correct IVector. 
    */

    if ( getAttribute(DataBlock->header.state) == STATE_POOL_NOT_INITIALIZED )
    {
        ShredData( iVector, IVECTOR_LENGTH);
        return ERROR_INCORRECT_POOL;
    }

    /*	Load the Key into the Algorithm driver and again, 
    the Pool is filled by zeros, because we are going 
    to decrypt random data (Pool stored in the DataBlock) */

    /* Load the Key to encryption algorithm module */
    if ( alg_make_key (Alg, Key, AlgKeyLength, PoolBuffer, &keyHandle ) )
    {
        ShredData( iVector, IVECTOR_LENGTH);
        return ERROR_INVALID_ALGORITHM;
    }

    /* copy encrypted pool from DataBlock to the PoolBuffer */
    memcpy( PoolBuffer, DataBlock->pool, POOL_SIZE_BYTES );

    /* Decrypt pool */
    if ( alg_decrypt(    Alg, keyHandle, iVector, 
                         PoolBuffer, POOL_SIZE_BYTES ) )
    {
        alg_free_key( Alg, keyHandle );
        ShredData( iVector, IVECTOR_LENGTH);
        return ERROR_INVALID_ALGORITHM;
    }

    /* Erase and free the Key data from the encryption algorithm module */
    ShredData( iVector, IVECTOR_LENGTH);
    alg_free_key( Alg, keyHandle );

    return ERROR_NO;
}



DWORD DataBlockChangePassword(
                             DATA_BLOCK *DataBlock,
                             ALG_SERV Alg,
                             DWORD      AlgKeyLength,
                             char *    Password,
                             char *    NewPassword,
                             BYTE       *PoolBuffer
                             )
{
    KEY_BLOCK *kb;
    BYTE  storageKey[ MAXIMUM_KEY_SIZE_BYTES ];
    BYTE  keyAndDigest[ MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES ];
    BYTE  calculatedDigest[ MAXIMUM_DIGEST_SIZE_BYTES ];
    DWORD keyHandle;
    BYTE  iVector[ IVECTOR_LENGTH ];
    BOOL  result = FALSE;

    kb = DataBlock->keys;

    /*  we fill the PoolBuffer with zeros, because 
        the buffers in the DataBlock we are going to decrypt 
        contain random data */

    memset( PoolBuffer, 0, POOL_SIZE_BYTES );

    for ( int i=0; i<MAXIMUM_NUMBER_OF_KEYS; i++, kb++ )
    {
        if ( getAttribute(kb->keyAttribute) == KATTRIBUTE_KEY_EMPTY ) continue;

        /* Calculate encryption key for decrypting KEY_BLOCK */
        if ( !GetStorageKeyFromPassword( Password, storageKey ) )
            return ERROR_INVALID_ALGORITHM;

        /* Decrypt KEY_BLOCK kb */
        /* First, load storageKey into the Algorithm driver */
        if ( alg_make_key( Alg, storageKey, AlgKeyLength, PoolBuffer, &keyHandle ) )
            return ERROR_INVALID_ALGORITHM;

        /* Now we don't need in storageKey, because it is copied to the
           Algorithm Driver */
        ShredData( storageKey, MAXIMUM_KEY_SIZE_BYTES );

        /* decrypt the key and its digest copied to the KEY_BLOCK */
        memcpy( keyAndDigest, kb->key, MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES );

        /* initialize iVector with any value.
           After first decrypting we will define it exactly */
        for ( i=0; i<IVECTOR_LENGTH; i++ ) iVector[i] = 0;

        /* first decryption */
        if ( alg_decrypt( Alg, keyHandle, iVector, 
                          keyAndDigest, MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES) )
        {
            alg_free_key( Alg, keyHandle );
            ShredData( keyAndDigest, MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES);
            return ERROR_INVALID_ALGORITHM;
        }

        /* define Initial Vector */
        memcpy( iVector, &(keyAndDigest[MAXIMUM_KEY_SIZE_BYTES]), IVECTOR_LENGTH );

        /* second decryption with exact Initial Vector */
        memcpy( keyAndDigest, kb->key, MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES );

        if ( alg_decrypt( Alg, keyHandle, iVector, 
                          keyAndDigest, MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES) )
        {
            alg_free_key( Alg, keyHandle );
            ShredData( keyAndDigest, MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES);
            return ERROR_INVALID_ALGORITHM;
        }

        /* erase storageKey from the Algorithm driver memory */
        alg_free_key( Alg, keyHandle );

        /* calculate digest of the key and compare it with digest */
        if ( !CalculateDigest( keyAndDigest, MAXIMUM_KEY_SIZE_BYTES, 
                               calculatedDigest ) )
        {
            ShredData( keyAndDigest, MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES);
            return ERROR_INVALID_ALGORITHM;
        }

        if ( memcmp( calculatedDigest, 
                     &(keyAndDigest[MAXIMUM_KEY_SIZE_BYTES]), 
                     SHA256_DIGEST_SIZE) == 0 )
        {
            /* Check if Alternative Key Block is initialized. If no, initialize it */
            createAlternativeKeyBlock(  DataBlock, Alg, AlgKeyLength, keyAndDigest, PoolBuffer );

            /*	Copy decrypted key to the KEY_BLOCK, then we will 
                encrypt it using new password */
            memcpy( kb->key, keyAndDigest, MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES);

            /*	We found the KEY_BLOCK encrypted with password,
                now we will encrypt it using newPassword
                First of all, calculate storageKey from the newPassword
             */
            if ( !GetStorageKeyFromPassword( NewPassword, storageKey ) )
                return ERROR_INVALID_ALGORITHM;

            /* Load storageKey into the Algorithm driver */
            if ( alg_make_key( Alg, storageKey, AlgKeyLength, PoolBuffer, &keyHandle ) )
            {
                ShredData( storageKey, MAXIMUM_KEY_SIZE_BYTES );
                return ERROR_INVALID_ALGORITHM;
            }

            /* Now we don't need in storageKey, because it is copied to the
               Algorithm Driver */
            ShredData( storageKey, MAXIMUM_KEY_SIZE_BYTES );

            /* encrypt the key and its digest in the DataBlock */
            if ( alg_encrypt( Alg, keyHandle, iVector, 
                              kb->key, MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES) )
            {
                alg_free_key( Alg, keyHandle );
                return ERROR_INVALID_ALGORITHM;
            }

            /* erase storageKey from the Algorithm driver memory */
            alg_free_key( Alg, keyHandle );
            result = TRUE;  
        }
    }

    if ( !result ) return ERROR_INCORRECT_PASSWORD;

    /* DataBlock is changed, update DataBlock's digest */
    if ( !DataBlockUpdateDigest( DataBlock ) )
        return ERROR_INTERNAL_PROBLEM;

    return ERROR_NO;
}


DWORD GenerateKey(  BYTE *Key,
                    int   KeyLength,
                    BYTE *Seed,
                    int   SeedLength )
{
    int k, s, i, kReminder;

    k = KeyLength  / SHA256_DIGEST_SIZE;
    s = SeedLength / k;

    if ( s >= SHA256_DIGEST_SIZE )
    {
        for ( i=0; i < k; i++ )
        {
            if ( i == (k-1) ) kReminder = KeyLength - k*SHA256_DIGEST_SIZE;
            else            kReminder = 0;

            if ( SHA256RandomGenerate(
                                     &(Key[ SHA256_DIGEST_SIZE * i ]),
                                     SHA256_DIGEST_SIZE + kReminder, 
                                     &(Seed[ s*i ]), 
                                     SeedLength - s*i) != ERROR_NO )
                return ERROR_INTERNAL_PROBLEM;
        }
    } else
    {
        if ( SHA256RandomGenerate(Key, KeyLength, Seed, SeedLength) != ERROR_NO )
            return ERROR_INTERNAL_PROBLEM;
    }

    return ERROR_NO;
}



/**********************************************************************
 * 
 * Local useful procedures
 *
 **********************************************************************/

DWORD getAttribute( DWORD AttBuffer )
{
    return(AttBuffer & (~KATTRIBUTE_MASK));
}

void setAttribute( DWORD *AttBuffer, DWORD Attribute )
{
    // Not all bites in the AttBuffer are used: 
    // in the current version only 2 bites from 32 are valid.
    // Because all 32 bits was previously initialized by random 
    // data, let unused bits continue to be random to avoid 
    // the known-plaintext attack.

    *AttBuffer &= KATTRIBUTE_MASK;
    *AttBuffer |= Attribute;
}


BOOL CalculateDigest( BYTE *Data, DWORD Length, BYTE *Digest )
{
    if (SHA256_MakeDigest( Data, (int)Length, Digest ) != SHA256_ERROR_NO)
        return FALSE;

    return TRUE;
}



BOOL DataBlockVerifyDigest( DATA_BLOCK *DataBlock )
{
    BYTE digest[ MAXIMUM_DIGEST_SIZE_BYTES ];

    if ( !CalculateDigest( (BYTE *)DataBlock, 
                           sizeof(DB_HEADER) + sizeof(KEY_BLOCK)*MAXIMUM_NUMBER_OF_KEYS,
                           digest) )
        return FALSE;

    if ( memcmp( DataBlock->digest, digest, SHA256_DIGEST_SIZE ) != 0 )
        return FALSE;

    return TRUE;
}



BOOL DataBlockUpdateDigest( DATA_BLOCK *DataBlock )
{
    BYTE digest[ MAXIMUM_DIGEST_SIZE_BYTES ];

    if ( !CalculateDigest( (BYTE *)DataBlock, 
                           sizeof(DB_HEADER) + sizeof(KEY_BLOCK)*MAXIMUM_NUMBER_OF_KEYS,
                           digest) )
        return FALSE;

    memcpy( DataBlock->digest, digest, SHA256_DIGEST_SIZE );

    return TRUE;
}


BOOL GetStorageKeyFromPassword( char *Password, BYTE *StorageKey )
{
    if ( ((DWORD)SHA256RandomGenerate( StorageKey, MAXIMUM_KEY_SIZE_BYTES, 
                                       (BYTE *)Password, strlen(Password))) != ERROR_NO )
        return FALSE;

    return TRUE;
}

//________________________________________________________________


BOOL createAlternativeKeyBlock(
                              DATA_BLOCK *DataBlock,
                              ALG_SERV   Alg,
                              DWORD      AlgKeyLength,
                              BYTE       *Key,
                              BYTE       *PoolBuffer
                              )
{
    DWORD                 keyHandle, i;
    KEY_BLOCK             *kb;
    KEY_BLOCK_ALTERNATIVE *kbFound;
    BYTE                  iVector[ IVECTOR_LENGTH ];

    /* check if Alternative Key Block already exists */
    kb = DataBlock->keys;

    for ( i=0; i<MAXIMUM_NUMBER_OF_KEYS; i++, kb++ )
    {
        if ( getAttribute(kb->keyAttribute) == KATTRIBUTE_ALTERNATIVE_BLOCK )
            return TRUE;
    }

    /* find empty KEY_BLOCK inside DataBlock*/
    kb = DataBlock->keys;
    kbFound = NULL;

    for ( i=0; i<MAXIMUM_NUMBER_OF_KEYS; i++, kb++ )
    {
        if ( getAttribute(kb->keyAttribute) == KATTRIBUTE_KEY_EMPTY )
        {
            kbFound = (KEY_BLOCK_ALTERNATIVE *)kb;
            break;
        }
    }

    if ( !kbFound ) return FALSE;

    // Empty key block is already filled in by random data.
    // We'll create digest for the data and encrypt it using the Key

    if ( !CalculateDigest( kbFound->verifyBytes, VERIFY_BYTES_LENGTH, kbFound->digest ) )
        return FALSE;

    memset( PoolBuffer, 0, POOL_SIZE_BYTES );
    memset( iVector,    0, IVECTOR_LENGTH ); 

    /* Load the Key to encryption algorithm module */
    if ( alg_make_key( Alg, Key, AlgKeyLength, PoolBuffer, &keyHandle ) )
        return FALSE;

    /* Encrypt 'verify data' and 'digest' */
    if ( alg_encrypt(    Alg, keyHandle, iVector, 
                         kbFound->verifyBytes, VERIFY_BYTES_LENGTH + MAXIMUM_DIGEST_SIZE_BYTES ) )
    {
        alg_free_key( Alg, keyHandle );
        return FALSE;
    }

    /* Erase and free the Key data from the encryption algorithm module */
    alg_free_key( Alg, keyHandle );

    setAttribute( &(kbFound->keyAttribute), KATTRIBUTE_ALTERNATIVE_BLOCK );

    return TRUE;
}




/**** Procedures for creating and opening Hidden part of container ****/

/**********************************************************************
 * 
 * DataBlockWriteKey_Hidden() .B- 
 *   fills a LAST free KEY_BLOCK in the DataBlock without 
 *   changing the KATTRIBUTE_KEY_EMPTY attribute
 *
 **********************************************************************/

#define OFFSET_IN_KEY_SPACE  (MAXIMUM_KEY_SIZE_BYTES - sizeof(DWORD) - sizeof(DWORD))
#define KEYMAP_IN_KEY_SPACE  (MAXIMUM_KEY_SIZE_BYTES - sizeof(DWORD))


static DWORD localDataBlockWriteKey_Hidden(
                                          DATA_BLOCK *DataBlock,
                                          KEY_BLOCK  *kbFound,
                                          ALG_SERV    Alg,
                                          DWORD      AlgId,
                                          DWORD      AlgKeyLength,
                                          char       *Password,
                                          BYTE       *Key,
                                          DWORD      KeyAttribute, // in the current release we don't use the parameter
                                          BYTE       *PoolBuffer,
                                          DWORD      Offset,
                                          DWORD      Mask
                                          )
{
    BYTE       storageKey[ MAXIMUM_KEY_SIZE_BYTES ];
    DWORD      keyHandle, *offsetPlace;
    BYTE       iVector[ IVECTOR_LENGTH ];

    /* copy encryption key to the KEY_BLOCK memory */
    memcpy( kbFound->key, Key, MAXIMUM_KEY_SIZE_BYTES );

    /* save offset inside last 4 bytes of key space.
       we reserved 64 bytes for key space,o last 4 bytes 
       are not busy by encryption key - usually up to 32 bytes, maximum one is Blowfish-448,
       hence in any case we have 64 bits in key space free. Will use the space
       for Offset and Mask)
     */
    offsetPlace = (DWORD *)(&( kbFound->key[OFFSET_IN_KEY_SPACE] ));
    *offsetPlace = Offset;

    offsetPlace = (DWORD *)(&( kbFound->key[KEYMAP_IN_KEY_SPACE] ));
    *offsetPlace = Mask;

    /* calculate digest of the key and save it in KEY_BLOCK */
    if ( !CalculateDigest( kbFound->key, MAXIMUM_KEY_SIZE_BYTES - sizeof(DWORD), 
                           kbFound->digest ) )
    {
        ShredData( kbFound->key, MAXIMUM_KEY_SIZE_BYTES );
        return ERROR_INVALID_ALGORITHM;
    }

    memcpy( iVector, kbFound->digest, IVECTOR_LENGTH );

    /* We don't encrypt pool in the data block with the Key,
       because it's already encrypted.
    */

    /*	Encrypt file-container's encryption key 
        and its digest using storageKey.
        First of all, calculate storageKey from the password
     */
    if ( !GetStorageKeyFromPassword( Password, storageKey ) )
    {
        ShredData( kbFound->key, MAXIMUM_KEY_SIZE_BYTES );
        ShredData( kbFound->digest, MAXIMUM_DIGEST_SIZE_BYTES );
        ShredData( iVector, IVECTOR_LENGTH);
        return ERROR_INVALID_ALGORITHM;
    }

    /*	Load storageKey into the Algorithm driver and again, 
        the Pool must be filled by zeros, because we are going 
        to encrypt random data (Key) */

    memset( PoolBuffer, 0, POOL_SIZE_BYTES );

    if ( alg_make_key( Alg, storageKey, AlgKeyLength, 
                       PoolBuffer, &keyHandle ) )
    {
        ShredData( storageKey, MAXIMUM_KEY_SIZE_BYTES );
        ShredData( kbFound->key, MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES);
        ShredData( iVector, IVECTOR_LENGTH);
        return ERROR_INVALID_ALGORITHM;
    }

    /* Now we don't need in storageKey, because it is copied to the
       Algorithm Driver */
    ShredData( storageKey, MAXIMUM_KEY_SIZE_BYTES );

    /* encrypt the key and its digest copied to the KEY_BLOCK */
    if ( alg_encrypt( Alg, keyHandle, iVector, 
                      kbFound->key, MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES) )
    {
        alg_free_key( Alg, keyHandle );
        ShredData( kbFound->key, MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES);
        ShredData( iVector, IVECTOR_LENGTH);
        return ERROR_INVALID_ALGORITHM;
    }

    /* erase storageKey from the Algorithm driver memory */
    alg_free_key( Alg, keyHandle );

    ShredData( iVector, IVECTOR_LENGTH);

    if ( !DataBlockUpdateDigest( DataBlock ) )
    {
        return ERROR_INTERNAL_PROBLEM;
    }

    return ERROR_NO;
}


DWORD DataBlockWriteKey_Hidden(
                              DATA_BLOCK *DataBlock,
                              ALG_SERV   Alg,
                              DWORD      AlgKeyLength,
                              char       *Password,
                              BYTE       *Key,
                              DWORD      KeyAttribute, // in the current release we don't use the parameter
                              BYTE       *PoolBuffer,
                              DWORD      Offset,
                              DWORD      *BusyMask,
                              DWORD      KeyBlockPositionMask
                              )
{
    KEY_BLOCK *kbFound;
    int       i;
    DWORD     mask, dwRes, position;

    dwRes = ERROR_NOT_ENOUGH_SPACE_FOR_KEY;

    /* Check if Alternative Key Block is initialized. If no, initialize it */
    createAlternativeKeyBlock(  DataBlock, Alg, AlgKeyLength, Key, PoolBuffer );

    if ( KeyBlockPositionMask == HIDDEN_POSITION_ANY )
    {
        for ( i = MAXIMUM_NUMBER_OF_KEYS - 1;  i>0;  i-- )
        {
            kbFound = &(DataBlock->keys[i]);
            if ( getAttribute(kbFound->keyAttribute) != KATTRIBUTE_KEY_EMPTY ) continue;

            if ( (*BusyMask) & (1 << i) ) continue;

            mask = (*BusyMask) | (1 << i);

            dwRes = localDataBlockWriteKey_Hidden( DataBlock, kbFound, Alg, Alg.alg_id, AlgKeyLength,
                                                   Password, Key, KeyAttribute, PoolBuffer, Offset, mask );

            if ( dwRes == ERROR_NO )
            {
                *BusyMask = mask;
                return ERROR_NO;
            }
        }
    } else
    {
        position = KeyBlockPositionMask;

        for ( i = 0;  (i<MAXIMUM_NUMBER_OF_KEYS) && (position != 1);  i++ )
        {
            position >>= 1;
        }

        if ( i==MAXIMUM_NUMBER_OF_KEYS ) return ERROR_INTERNAL_PROBLEM;

        kbFound = &(DataBlock->keys[i]);

        if ( getAttribute(kbFound->keyAttribute) != KATTRIBUTE_KEY_EMPTY )
            return ERROR_INTERNAL_PROBLEM;

        mask = (*BusyMask) | (1 << i);

        dwRes = localDataBlockWriteKey_Hidden( DataBlock, kbFound, Alg, Alg.alg_id, AlgKeyLength,
                                               Password, Key, KeyAttribute, PoolBuffer, Offset, mask );

        if ( dwRes == ERROR_NO )
        {
            *BusyMask = mask;
            return ERROR_NO;
        }
    }

    return dwRes;
}


/*******************************************************
 *
 * DataBlockVerifyPasswordAndGetKey_Hidden()
 *
 *******************************************************/

static DWORD localDataBlockVerifyPasswordAndGetKey_Hidden(
                                                         DATA_BLOCK *DataBlock,
                                                         KEY_BLOCK  *kbFound,
                                                         ALG_SERV   Alg,
                                                         DWORD      AlgId,
                                                         DWORD      AlgKeyLength,
                                                         char       *Password,
                                                         BYTE       *Key,
                                                         BYTE       *PoolBuffer,
                                                         DWORD      *Offset,
                                                         DWORD      *BusyMask
                                                         )
{
    BYTE  storageKey[ MAXIMUM_KEY_SIZE_BYTES ];
    BYTE  keyAndDigest[ MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES ];
    BYTE  calculatedDigest[ MAXIMUM_DIGEST_SIZE_BYTES ];
    DWORD keyHandle;
    BYTE  iVector[ IVECTOR_LENGTH ];

    memset( PoolBuffer, 0, POOL_SIZE_BYTES );

    if ( getAttribute(kbFound->keyAttribute) != KATTRIBUTE_KEY_EMPTY )
        return ERROR_INCORRECT_PASSWORD;

    /* Calculate encryption key for decrypting KEY_BLOCK */
    if ( !GetStorageKeyFromPassword( Password, storageKey ) )
        return ERROR_INVALID_ALGORITHM;

    /* Decrypt KEY_BLOCK kb */

    /* First, load storageKey into the Algorithm driver */
    if ( alg_make_key( Alg, storageKey, AlgKeyLength, 
                       PoolBuffer, &keyHandle ) )
        return ERROR_INVALID_ALGORITHM;

    /* Now we don't need in storageKey, because it is saved in the Algorithm Driver */
    ShredData( storageKey, MAXIMUM_KEY_SIZE_BYTES );

    /* decrypt the key and its digest copied to the KEY_BLOCK */
    memcpy( keyAndDigest, kbFound->key, MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES );

    /*	initialize iVector with any value.
        After first decrypting we will define it exactly */
    memset( iVector, 0, IVECTOR_LENGTH );

    /* first decryption */
    if ( alg_decrypt( Alg, keyHandle, iVector, 
                      keyAndDigest, MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES) )
    {
        alg_free_key( Alg, keyHandle );
        ShredData( keyAndDigest, MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES);
        return ERROR_INVALID_ALGORITHM;
    }

    /* define Initial Vector exactly */
    memcpy( iVector, &(keyAndDigest[MAXIMUM_KEY_SIZE_BYTES]), IVECTOR_LENGTH );

    /* second decryption with exact Initial Vector */
    memcpy( keyAndDigest, kbFound->key, MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES );

    if ( alg_decrypt( Alg, keyHandle, iVector, 
                      keyAndDigest, MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES) )
    {
        alg_free_key( Alg, keyHandle );
        ShredData( keyAndDigest, MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES);
        ShredData( iVector, IVECTOR_LENGTH);
        return ERROR_INVALID_ALGORITHM;
    }

    /* erase storageKey from the Algorithm driver memory */
    alg_free_key( Alg, keyHandle );

    /* calculate digest of the key and compare it with digest */
    if ( !CalculateDigest( keyAndDigest, MAXIMUM_KEY_SIZE_BYTES - sizeof(DWORD), 
                           calculatedDigest ) )
    {
        ShredData( keyAndDigest, MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES);
        ShredData( iVector, IVECTOR_LENGTH);
        return ERROR_INVALID_ALGORITHM;
    }

    if ( memcmp( calculatedDigest, &(keyAndDigest[MAXIMUM_KEY_SIZE_BYTES]), SHA256_DIGEST_SIZE) != 0 )
    {
        ShredData( iVector, IVECTOR_LENGTH);
        return ERROR_INCORRECT_PASSWORD;
    }

    /* Get encryption key */
    memcpy( Key, keyAndDigest, MAXIMUM_KEY_SIZE_BYTES );

    /* Get offset of the hidden part inside container */
    *Offset = *(DWORD*)(keyAndDigest + OFFSET_IN_KEY_SPACE );

    *BusyMask = *(DWORD*)(keyAndDigest + KEYMAP_IN_KEY_SPACE );

    /*	In the following part we will decrypt and place a correct 
        pool data to the PoolBuffer,
        Now everything is Ok, and we have correct IVector. 
    */
    if ( getAttribute(DataBlock->header.state) == STATE_POOL_NOT_INITIALIZED )
    {
        ShredData( iVector, IVECTOR_LENGTH);
        return ERROR_INCORRECT_POOL;
    }

    /*	Load the Key into the Algorithm driver and again, 
    the Pool is filled by zeros, because we are going 
    to decrypt random data (Pool stored in the DataBlock) */

    /* Load the Key to encryption algorithm module */
    if ( alg_make_key( Alg, Key, AlgKeyLength, 
                       PoolBuffer, &keyHandle ) )
    {
        ShredData( iVector, IVECTOR_LENGTH);
        return ERROR_INVALID_ALGORITHM;
    }

    /* copy encrypted pool from DataBlock to the PoolBuffer */
    memcpy( PoolBuffer, DataBlock->pool, POOL_SIZE_BYTES );

    /* Decrypt pool */
    if ( alg_decrypt(    Alg, keyHandle, iVector, 
                         PoolBuffer, POOL_SIZE_BYTES ) )
    {
        alg_free_key( Alg, keyHandle );
        ShredData( iVector, IVECTOR_LENGTH);
        return ERROR_INVALID_ALGORITHM;
    }

    /* Erase and free the Key data from the encryption algorithm module */
    ShredData( iVector, IVECTOR_LENGTH);
    alg_free_key( Alg, keyHandle );

    return ERROR_NO;
}



DWORD DataBlockVerifyPasswordAndGetKey_Hidden(
                                             DATA_BLOCK *DataBlock,
                                             ALG_SERV   Alg,
                                             DWORD      AlgKeyLength,
                                             char       *Password,
                                             BYTE       *Key,
                                             BYTE       *PoolBuffer,
                                             DWORD      *Offset,
                                             DWORD      *BusyMask,
                                             DWORD      *KeyBlockPositionMask
                                             )
{
    KEY_BLOCK *kbFound;
    int       i;
    DWORD     dwRes;

    dwRes = ERROR_INCORRECT_PASSWORD;

    for ( i = MAXIMUM_NUMBER_OF_KEYS - 1;  i>0;  i-- )
    {
        kbFound = &(DataBlock->keys[i]);
        if ( getAttribute(kbFound->keyAttribute) != KATTRIBUTE_KEY_EMPTY ) continue;

        dwRes = localDataBlockVerifyPasswordAndGetKey_Hidden( DataBlock, kbFound,
                                                              Alg, Alg.alg_id, AlgKeyLength, Password, Key,
                                                              PoolBuffer, Offset, BusyMask );

        if ( dwRes == ERROR_NO )
        {
            *KeyBlockPositionMask = 1 << i;
            return ERROR_NO;
        }
    }

    return dwRes;
}



/*******************************************************
 *
 * DataBlock_CheckFreeSpaceForNewPassword()
 * - check - can we add new password to container?
 *
 *******************************************************/


DWORD DataBlock_CheckFreeSpaceForNewPassword( DATA_BLOCK *DataBlock )
{
    KEY_BLOCK *kb, *kbFound;

    /* find empty KEY_BLOCK inside DataBlock */
    kb = DataBlock->keys;
    kbFound = NULL;

    for ( int i=0; i<MAXIMUM_NUMBER_OF_KEYS; i++, kb++ )
    {
        if ( getAttribute(kb->keyAttribute) == KATTRIBUTE_KEY_EMPTY )
        {
            kbFound = kb;
            break;
        }
    }

    if ( !kbFound ) return ERROR_NOT_ENOUGH_SPACE_FOR_KEY;

    return ERROR_NO;
}



/*******************************************************
 *
 * DataBlockAddPassword()
 * - add new password to container
 *
 *******************************************************/


DWORD DataBlockAddPassword( DATA_BLOCK **DataBlock,
                            DWORD      *DataSize,
                            ALG_SERV   Alg,
                            DWORD      AlgKeyLength,
                            char       *Password,
                            BYTE       *Key,
                            DWORD      KeyAttribute,
                            BYTE       *PoolBuffer
                          )
{
    BYTE      storageKey[ MAXIMUM_KEY_SIZE_BYTES ];
    DWORD     keyHandle;
    KEY_BLOCK *kb, *kbFound;
    BYTE iVector[ IVECTOR_LENGTH ];

    /* find empty KEY_BLOCK inside DataBlock */
    kb = (*DataBlock)->keys;
    kbFound = NULL;

    for ( int i=0; i<MAXIMUM_NUMBER_OF_KEYS; i++, kb++ )
    {
        if ( getAttribute(kb->keyAttribute) == KATTRIBUTE_KEY_EMPTY )
        {
            kbFound = kb;
            break;
        }
    }

    memset( PoolBuffer, 0, POOL_SIZE_BYTES );

    if ( !kbFound ) return ERROR_NOT_ENOUGH_SPACE_FOR_KEY;

    /* copy encryption key to the KEY_BLOCK memory */
    memcpy( kbFound->key, Key, MAXIMUM_KEY_SIZE_BYTES );

    /* calculate digest of the key and save it in KEY_BLOCK */
    if ( !CalculateDigest( kbFound->key, MAXIMUM_KEY_SIZE_BYTES, 
                           kbFound->digest ) )
    {
        ShredData( kbFound->key, MAXIMUM_KEY_SIZE_BYTES );
        return ERROR_INVALID_ALGORITHM;
    }

    memcpy( iVector, kbFound->digest, IVECTOR_LENGTH );

    /*	We begin to encrypt file-container's encryption key 
        and its digest using storageKey.
        First of all, calculate storageKey from the password
     */
    if ( !GetStorageKeyFromPassword( Password, storageKey ) )
    {
        ShredData( kbFound->key, MAXIMUM_KEY_SIZE_BYTES );
        ShredData( kbFound->digest, MAXIMUM_DIGEST_SIZE_BYTES );
        ShredData( iVector, IVECTOR_LENGTH);
        return ERROR_INVALID_ALGORITHM;
    }

    /*	Load storageKey into the Algorithm driver and again, 
        the Pool must be filled by zeros, because we are going 
        to encrypt random data (Key) */
    if ( alg_make_key( Alg, storageKey, AlgKeyLength, 
                       PoolBuffer, &keyHandle ) )
    {
        ShredData( storageKey, MAXIMUM_KEY_SIZE_BYTES );
        ShredData( kbFound->key, MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES);
        ShredData( iVector, IVECTOR_LENGTH);
        return ERROR_INVALID_ALGORITHM;
    }

    /* Now we don't need in storageKey, because it is copied to the
       Algorithm Driver */
    ShredData( storageKey, MAXIMUM_KEY_SIZE_BYTES );

    /* encrypt the key and its digest copied to the KEY_BLOCK */
    if ( alg_encrypt( Alg, keyHandle, iVector, 
                      kbFound->key, MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES) )
    {
        alg_free_key( Alg, keyHandle );
        ShredData( kbFound->key, MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES);
        ShredData( iVector, IVECTOR_LENGTH);
        return ERROR_INVALID_ALGORITHM;
    }

    /* erase storageKey from the Algorithm driver memory */
    alg_free_key( Alg, keyHandle );

    ShredData( iVector, IVECTOR_LENGTH);

    setAttribute( &(kbFound->keyAttribute), KeyAttribute );

    /* Check if Alternative Key Block is initialized. If no, initialize it */
    createAlternativeKeyBlock(  *DataBlock, Alg, AlgKeyLength, Key, PoolBuffer );

    if ( !DataBlockUpdateDigest( *DataBlock ) )
    {
        setAttribute( &(kbFound->keyAttribute), KATTRIBUTE_KEY_EMPTY );
        return ERROR_INTERNAL_PROBLEM;
    }

    return ERROR_NO;
}




/*******************************************************
 *
 * DataBlockAddPassword_Hidden()
 * - add new password for hidden part of container
 *
 *******************************************************/


DWORD DataBlockAddPassword_Hidden(
                                 DATA_BLOCK **DataBlock,
                                 DWORD      *DataSize,
                                 ALG_SERV    Alg,
                                 DWORD      AlgKeyLength,
                                 char       *Password,
                                 BYTE       *Key,
                                 DWORD      KeyAttribute, // in the current release we don't use the parameter
                                 BYTE       *PoolBuffer,
                                 DWORD      Offset,
                                 DWORD      *BusyMask
                                 )
{
    createAlternativeKeyBlock( *DataBlock, Alg, AlgKeyLength, Key, PoolBuffer );

    return DataBlockWriteKey_Hidden( *DataBlock, Alg, AlgKeyLength,
                                     Password, Key, KeyAttribute, PoolBuffer, 
                                     Offset, BusyMask, HIDDEN_POSITION_ANY );
}








/*******************************************************
 *
 * DataBlockRemoveAdditionalPassword()
 * - removes additional password for container
 *
 *******************************************************/

DWORD DataBlockRemoveAdditionalPassword(
                                       DATA_BLOCK *DataBlock,
                                       ALG_SERV   Alg,
                                       DWORD      AlgKeyLength,
                                       char       *Password,
                                       BYTE       *Key,
                                       BYTE       *PoolBuffer
                                       )
{
    KEY_BLOCK *kb, *kbFound;
    BYTE  storageKey[ MAXIMUM_KEY_SIZE_BYTES ];
    BYTE  keyAndDigest[ MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES ];
    BYTE  calculatedDigest[ MAXIMUM_DIGEST_SIZE_BYTES ];
    DWORD keyHandle, i;
    BYTE  iVector[ IVECTOR_LENGTH ];

    kb = DataBlock->keys;
    kbFound = NULL;

    /*  we fill the PoolBuffer with zeros, because 
        the buffers in the DataBlock we are going to decrypt 
        contain random data */

    /*	in the first part of the procedure we use the PoolBuffer
        to initialize encryption algorithm module with zeroed pools.
        In the second part we will place a correct decrypted pool 
        data to the PoolBuffer */

    memset( PoolBuffer, 0, POOL_SIZE_BYTES );

    for ( i=0; i<MAXIMUM_NUMBER_OF_KEYS; i++, kb++ )
    {
        if ( getAttribute(kb->keyAttribute) == KATTRIBUTE_KEY_EMPTY ) continue;

        /* Calculate encryption key for decrypting KEY_BLOCK */
        if ( !GetStorageKeyFromPassword( Password, storageKey ) )
            return ERROR_INVALID_ALGORITHM;

        /* Decrypt KEY_BLOCK kb */
        /* First, load storageKey into the Algorithm driver */
        if ( alg_make_key( Alg, storageKey, AlgKeyLength, 
                           PoolBuffer, &keyHandle ) )
            return ERROR_INVALID_ALGORITHM;

        /* Now we don't need in storageKey, because it is copied to the
           Algorithm Driver */
        ShredData( storageKey, MAXIMUM_KEY_SIZE_BYTES );

        /* decrypt the key and its digest copied to the KEY_BLOCK */
        memcpy( keyAndDigest, kb->key, MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES );

        /* initialize iVector with any value.
           After first decrypting we will define it exactly */
        memset( iVector, 0, IVECTOR_LENGTH );

        /* first decryption */
        if ( alg_decrypt( Alg, keyHandle, iVector, 
                          keyAndDigest, MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES) )
        {
            alg_free_key( Alg, keyHandle );
            ShredData( keyAndDigest, MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES);
            return ERROR_INVALID_ALGORITHM;
        }

        /* define Initial Vector */
        memcpy( iVector, &(keyAndDigest[MAXIMUM_KEY_SIZE_BYTES]), IVECTOR_LENGTH );

        /* second decryption with exact Initial Vector */
        memcpy( keyAndDigest, kb->key, MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES );

        if ( alg_decrypt( Alg, keyHandle, iVector, 
                          keyAndDigest, MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES) )
        {
            alg_free_key( Alg, keyHandle );
            ShredData( keyAndDigest, MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES);
            ShredData( iVector, IVECTOR_LENGTH);
            return ERROR_INVALID_ALGORITHM;
        }

        /* erase storageKey from the Algorithm driver memory */
        alg_free_key( Alg, keyHandle );

        /* calculate digest of the key and compare it with digest */
        if ( !CalculateDigest( keyAndDigest, MAXIMUM_KEY_SIZE_BYTES, 
                               calculatedDigest ) )
        {
            ShredData( keyAndDigest, MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES);
            ShredData( iVector, IVECTOR_LENGTH);
            return ERROR_INVALID_ALGORITHM;
        }

        if ( memcmp( calculatedDigest, 
                     &(keyAndDigest[MAXIMUM_KEY_SIZE_BYTES]), 
                     SHA256_DIGEST_SIZE) == 0 )
        {
            kbFound = kb;
            break;
        }
    }

    if ( !kbFound )
    {
        ShredData( iVector, IVECTOR_LENGTH);
        return ERROR_INCORRECT_PASSWORD;
    }

    ShredData( keyAndDigest, MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES);

    if ( i == 0 )
    {
        return ERROR_MASTER_PASSWORD_ENTERED;
    }

    memcpy( kbFound, keyAndDigest, MAXIMUM_KEY_SIZE_BYTES + MAXIMUM_DIGEST_SIZE_BYTES );

    kbFound->keyAttribute = KATTRIBUTE_KEY_EMPTY;

    memcpy( Key, keyAndDigest, MAXIMUM_KEY_SIZE_BYTES );

    if ( !DataBlockUpdateDigest( DataBlock ) )
    {
        return ERROR_INTERNAL_PROBLEM;
    }

    return ERROR_NO;
}



