/*******************************************************************
 *	Copyright (c) 1994-1998 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	File:		driver/bc_dev.c
 *
 *	Description:	BestCrypt pseudo-device driver.
 *
 *	Scope:		BestCrypt pseudo-device driver
 *
 *	Platforms:	Linux
 * 
 *	Ideas:		Igor M. Arsenin
 *			Serge A. Frolov
 *			Vitaliy G. Zolotarev
 *
 *	Author:		Nail R. Kaipov
 *
 *	Created:	10-Nov-1998
 *
 *	Revision:	06-Aug-1999.	new strategy routine
 *			15-Sep-1999.	2.3.x kernels support
 *			11-Oct-1999.	strategy bugfix
 *			15-Dec-1999.	driver refinement
 *			21-Mar-2000.	bugfix for large blocks (v0.4b-2)
 *			02-May-2001.	SMP bugfix, /proc update
 *			06-Jun-2001.	writing magic in bc_clr_fd
 *
 *******************************************************************/

#ifndef EXPORT_SYMTAB
#define EXPORT_SYMTAB
#endif

#include "bc_cfg.h"
#include <linux/module.h>
#include <linux/fs.h>			/* inode et al            */
#include <linux/proc_fs.h>		/* inode et al            */
#include <linux/vmalloc.h>		/* vmalloc()              */
#include <linux/kernel.h>		/* printk()               */
#include <asm/uaccess.h>		/* put_user()             */
#include <asm/atomic.h>			/* atomic operations      */
#include <linux/hdreg.h>		/* hd_geometry            */
#include <linux/file.h>			/* inode et al            */

#include "bc_types.h"			/* bc types               */
#include "bc_ioctl.h"			/*           ioctl codes  */
#include "bc_mgr.h"
#include "bc_dev.h"			/* bc types        codes  */

#define DEVICE_OFF(dev)
#include <linux/blk.h>

#define __KERNEL_SYSCALLS__
#include <linux/unistd.h>

static int              bc_devices = DEFAULT_BC_DEVICES;
static int		bc_partitions = DEFAULT_BC_PARTITIONS;
static int		bc_minor_shift = 0;
static int		bc_start_minor = 128;

static struct bc_device	*bc_dev;
static struct bc_disk	*bc_dsk;
static struct gendisk	bc_gendisk;
static int		bc_sizes	[MAX_MINORS];
static int		bc_blksizes	[MAX_MINORS];
static struct hd_struct bc_hd		[MAX_MINORS];
struct request		*BC_CURRENT = NULL;
static struct wait_queue *wait_open = 0;

/* pid control table support */
#define BC_PID_SIZE 64
static pid_t            	*bc_pid_table;
static int			bc_pid_size;
static int			bc_pid_next;
static struct semaphore 	bc_pid_sema;
static struct timer_list	bc_pid_timer;

struct bc_buffer_desc {
	atomic_t	count;
	char		*ptr;
} bc_buffers[BC_BUFFERS];

static ssize_t bc_read      (struct file *,   char *, size_t, loff_t *);
static ssize_t bc_write     (struct file *,   const char *, size_t, loff_t *);
static int     bc_ioctl     (struct inode *,  struct file *, unsigned int, unsigned long);
static int     bc_open      (struct inode *,  struct file *);	
static int     bc_release   (struct inode *,  struct file *);
static int     dummy_ioctl  (struct inode *i, struct file *f, u_int cmd, u_long arg) { return -ENODEV; }
static int     dummy_open   (struct inode *i, struct file *f) { return -ENODEV; }
static int     dummy_release(struct inode *i, struct file *f) { return -ENODEV; }
static int     *save_blksize;
static int     *save_blksize_size;
static int     max_loop;
static struct  proc_dir_entry    *proc_bcrypt, *proc_tmp;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,32))

#define BC_GET_REQUEST_FN(_MAJOR_) blk_dev[_MAJOR_].request_queue.request_fn
static void    dummy_request_fn  (request_queue_t *rq) { return; }
static void    (*save_request_fn)(request_queue_t *rq) = dummy_request_fn;
#else

#define BC_GET_REQUEST_FN(_MAJOR_) blk_dev[_MAJOR_].request_fn
static void    dummy_request_fn(void) { return; }
static void    (*save_request_fn)(void) = dummy_request_fn;
#endif

static struct file_operations save_fops = {
	read   :block_read,
	write  :block_write,
	ioctl  :dummy_ioctl,
	open   :dummy_open,
	release:dummy_release,
	fsync  :block_fsync
};

static struct file_operations bc_fops = {
	read   :bc_read,
	write  :bc_write,
	ioctl  :bc_ioctl,
	open   :bc_open,
	release:bc_release,
	fsync  :block_fsync
};

EXPORT_SYMBOL(register_bc_algo);
EXPORT_SYMBOL(unregister_bc_algo);

MODULE_PARM(bc_devices, "i");
MODULE_PARM_DESC(bc_devices, "Max. number of devices (1-127).");

MODULE_PARM(bc_partitions, "i");
MODULE_PARM_DESC(bc_partitions, "Max. number of partitions per device (0-31).");

MODULE_AUTHOR("Jetico, Inc.");
MODULE_DESCRIPTION("BestCrypt encrypted block device driver.");
/*-----------------------------------------------------------*/
static inline struct bc_device *get_bcdev(kdev_t dev)
{
	int	mn = MINOR(dev) - bc_start_minor;
	
	if ((MAJOR_NR == MAJOR(dev)) && (MAX_MINORS > MINOR(dev))) {
	    mn >>=bc_minor_shift;
	    if ((mn >= 0) && (mn < bc_devices))
		return bc_dev+mn;
	}
	return NULL;
}

static inline struct bc_disk *get_bcdsk(kdev_t dev)
{
	int	mn = MINOR(dev) - bc_start_minor;
	
	if ((MAJOR_NR == MAJOR(dev)) && (MAX_MINORS > MINOR(dev))) {
	    if ((mn >= 0) && (mn < bc_devices*bc_partitions))
		return bc_dsk+mn;
	}
	return NULL;
}

static inline int is_parent(struct bc_disk *bd)
{
	if (NULL != bd)
		return !((bd->bd_number-bc_start_minor) & ((1 << bc_minor_shift) - 1));
	return 0;
}

static inline void reset_dev(struct bc_device *bc)
{
	int            i, j;
	struct bc_disk *bd;
	
	if (NULL == bc)
		return;
		
	bc->bc_offset = 0L;
	bc->bc_start_sector = 0L;
	bc->bc_num_sectors  = 0L;
	bc->bc_dev = 0;
	bc->bc_key = 0;
	bc->bc_alg    = NULL;
	bc->bc_file   = NULL;
	bc->bc_dentry = NULL;
	bc->bc_activity = 0;
	memset(&bc->bc_flags, 0, sizeof(bc->bc_flags));
	
	bd = &bc_dsk[bc->bc_number << bc_minor_shift];
	for (i = 0; i < 1 << bc_minor_shift; i++) {
		j = bd->bd_number;
		memset(bd, 0, sizeof(struct bc_disk));
		bd->bd_number = j;
		bd++;
	}
	
}
/*-----------------------------------------------------------*/

static inline int bc_find_pid(pid_t pid) 
{
	register int i;
	for (i = 0; i < bc_pid_next; i++)
		if (pid == bc_pid_table[i])
			return i;
	return -1;
}

static int bc_find_pid_safe(pid_t pid)
{
	int i;
	down(&bc_pid_sema);
	i = bc_find_pid(pid);
	up(&bc_pid_sema);
	return i;
}
static int bc_add_pid(pid_t pid) 
{
	pid_t	*tmp;
	
	down(&bc_pid_sema);
	if (bc_find_pid(pid) > -1) {
		up(&bc_pid_sema);
		return 0;
	}
	if (bc_pid_next < bc_pid_size) {
		bc_pid_table[bc_pid_next] = pid;
		bc_pid_next++;
	} else {
		tmp = kmalloc((bc_pid_size+BC_PID_SIZE) * sizeof(pid_t), GFP_KERNEL);
		if (NULL == tmp) {
			up(&bc_pid_sema);
			return -ENOMEM;
		}
		memcpy(bc_pid_table, tmp, bc_pid_size * sizeof(pid_t));
		kfree(bc_pid_table);
		bc_pid_table = tmp;
		bc_pid_size += BC_PID_SIZE;
		bc_pid_table[bc_pid_next] = pid;
		bc_pid_next++;
	}
	if (1 == bc_pid_next) {
		bc_pid_timer.expires = jiffies + 5 * HZ;
		add_timer(&bc_pid_timer);
	}
	up(&bc_pid_sema);
	return 0;
}

static int bc_del_pid(pid_t pid) 
{
	int	i;
	pid_t	*tmp;

	down(&bc_pid_sema);

	i = bc_find_pid(pid);
	if (0 > i) {
		up(&bc_pid_sema);
		return 0;
	}
	bc_pid_next--;
	if (bc_pid_next)
	        bc_pid_table[i] = bc_pid_table[bc_pid_next];
	bc_pid_table[bc_pid_next] = 0;
		
	if (bc_pid_size-bc_pid_next > 2*BC_PID_SIZE) {
		tmp = kmalloc((bc_pid_size-BC_PID_SIZE) * sizeof(pid_t), GFP_KERNEL);
		if (NULL == tmp) {
			up(&bc_pid_sema);
			return 0;
		}
		bc_pid_size -= BC_PID_SIZE;
		memcpy(bc_pid_table, tmp, bc_pid_size * sizeof(pid_t));
		kfree(bc_pid_table);
		bc_pid_table = tmp;
	}
	up(&bc_pid_sema);
	return 0;
}

static void bc_pid_timer_proc(unsigned long unused) 
{
	register int i;
	
	down(&bc_pid_sema);
	for (i = 0; i < bc_pid_next; i++) {
		if (NULL == find_task_by_pid(bc_pid_table[i])) {
			bc_pid_next--;
			if (bc_pid_next)
				bc_pid_table[i] = bc_pid_table[bc_pid_next];
			bc_pid_table[bc_pid_next] = 0;
			i--;
		}
	}
	
	if (bc_pid_next) {
		bc_pid_timer.expires = jiffies + 5 * HZ;
		add_timer(&bc_pid_timer);
	}
	up(&bc_pid_sema);
}

/*-----------------------------------------------------------*/

static void grab_requests();
static void bc_request();

static void do_bc_request() 
{
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,32))
#define BC_NULL NULL
#else
#define BC_NULL
#endif
	if (NULL == CURRENT)
		return;
	grab_requests();
	save_request_fn(BC_NULL);
	bc_request(BC_NULL);
#undef BC_NULL
}


static void grab_requests()
{
	register struct request *prev, *tail, *curr;
	
	if (NULL == (curr = CURRENT))
		return;
	if (NULL != (tail = BC_CURRENT))
		while (tail->next)
			tail = tail->next;
	prev = NULL;
	while (curr) {
		if (get_bcdev(curr->rq_dev)) {
			if (tail) tail->next = curr;
			else      BC_CURRENT = curr;
			if (prev) prev->next = curr->next;
			else      CURRENT    = curr->next; 
			tail = curr;
		} else {
			if (prev) prev = prev->next;
			else      prev = CURRENT; 
		}
		curr = curr->next;
	}
	if (tail) tail->next = NULL;
}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,32))
static void bc_request(request_queue_t *abcd)
#else
static void bc_request()
#endif
{
	struct bc_device	*bc;
	struct bc_disk		*bd;
	struct request		*req;
	struct file		*file;
	size_t			len, ret, size;
	loff_t			pos, sec;
	char			*dst, *buf = NULL;
	u16			iv[4];
	u64			iv64, iv64le;
	mm_segment_t		fs;
	int			uptodate, i;

repeat:
	if (!BC_CURRENT) return;
	req = BC_CURRENT;
	BC_CURRENT = req->next;
	uptodate = 0;
	bc = get_bcdev(req->rq_dev);
	bd = get_bcdsk(req->rq_dev);
	if (!bc || !bd || !bc->bc_dentry || !bc->bc_dentry->d_inode) {
		goto error_out;
	}
	if (!bd->bd_flags.configured) {
		if (bc->bc_flags.busy) {
			bd->bd_offset = bc->bc_offset +
			(((loff_t)bc_gendisk.part[bd->bd_number].start_sect)<<9);
		} else {
			printk(KERN_ERR "bc: device not configured yet.\n");
			goto error_out;
		}
	}
	if (WRITE == req->cmd) {
		if (bc->bc_flags.readonly) {
			goto error_out;
		}
	} else if (READ != req->cmd) {
		printk(KERN_ERR "bc: unknown command (%d).", req->cmd);
		goto error_out;
	}
	sec   = (loff_t)req->sector + bc->bc_start_sector;
	pos   = sec * 512 + bd->bd_offset;
	len   = req->current_nr_sectors << 9;
	dst   = req->buffer;
	file  = bc->bc_file;
	iv[0] = iv[1] = iv[2] = iv[3] = sec & 0xFFFF;

	iv64  = sec+1;

	bc->bc_activity = 1;
	
	if (READ == req->cmd) {
		fs = get_fs();
		set_fs(KERNEL_DS);
		spin_unlock_irq(&io_request_lock);

		ret = file->f_op->read(file, dst, len, &pos);
		spin_lock_irq(&io_request_lock);
		set_fs(fs);
		if (ret != len) {
			printk(KERN_ERR "bc: read error wanted %ld, read %ld\n", 
				(unsigned long)len, (unsigned long)ret);
			goto error_out;
		}
		if (bc->bc_flags.iv_64bit) {
			iv64le = __cpu_to_le64(iv64);
			bc->bc_alg->decrypt(bc->bc_key, (char *)(&iv64le), dst, dst, len);
		} else {
			bc->bc_alg->decrypt(bc->bc_key, (char *)iv, dst, dst, len);
		}
	} else {
		for (i = 0; i < BC_BUFFERS; i++) {
			if (atomic_dec_and_test(&bc_buffers[i].count)) {
				buf = bc_buffers[i].ptr;
				break;
			} else
				atomic_inc(&bc_buffers[i].count);
		}
		if (BC_BUFFERS == i) {
			printk(KERN_ERR "bc: insufficient resources\n");
			goto error_out;
		}
		size = BC_BUFFER_SIZE;
		while (len > 0) {
			if (size > len)
				size = len;
			if (bc->bc_flags.iv_64bit) {
				iv64le = __cpu_to_le64(iv64);
				bc->bc_alg->encrypt(bc->bc_key, (char *)(&iv64le), dst, buf, size);
			} else {
				bc->bc_alg->encrypt(bc->bc_key, (char *)iv, dst, buf, size);
			}
			fs = get_fs();
			set_fs(KERNEL_DS);
			spin_unlock_irq(&io_request_lock);
			ret = file->f_op->write(file, buf, size, &pos);
			spin_lock_irq(&io_request_lock);
			set_fs(fs);
			if (ret != size) {
				printk(KERN_ERR "bc: write error wanted %ld, wrote %ld\n", 
					(unsigned long)size, (unsigned long)ret);
				atomic_inc(&bc_buffers[i].count);
				goto error_out;
			}
			if (bc->bc_flags.iv_64bit) {
				iv64 += BC_BUFFER_SIZE / 512;
			} else {
				iv[0] = ((int)iv[0] + BC_BUFFER_SIZE / 512) & 0xFFFF;
				iv[3] = iv[2] = iv[1] = iv[0];
			}
			dst += size;
			len -= size;
		}
		atomic_inc(&bc_buffers[i].count);
	}
	uptodate = 1;
error_out:	
        if (!end_that_request_first(req, uptodate, DEVICE_NAME))
		end_that_request_last(req);
	
	goto repeat;
}

/*-- end of strategy routine --------------------------------*/



/*-- fops functions -----------------------------------------*/
static int bc_open(struct inode *inode, struct file *file)
{
	struct bc_device *bc;

	if (!inode)	
		return -EINVAL;
	bc = get_bcdev(inode->i_rdev);
	if (NULL == bc)
		return save_fops.open(inode, file);

	MOD_INC_USE_COUNT;

	while (1) {
		down(&bc->bc_control);
		if (!bc->bc_flags.busy)
			break;
		up(&bc->bc_control);
		sleep_on(&wait_open);
	}
	bc->bc_refcnt++;
	up(&bc->bc_control);

	if (capable(CAP_SYS_ADMIN))
		bc_add_pid(current->pid);	

	return 0;
}

static int bc_release(struct inode *inode, struct file *file)
{
	struct bc_device *bc;
	
	if (!inode)
		return 0;

	bc = get_bcdev(inode->i_rdev);
	if (NULL == bc)
		return save_fops.release(inode, file);

	down(&bc->bc_control);
	if (bc->bc_refcnt <= 0)	
		printk(KERN_WARNING "bc_release: refcount (%d) <= 0\n", bc->bc_refcnt);
	else  {
		bc->bc_refcnt--;
		MOD_DEC_USE_COUNT;
	}
	up(&bc->bc_control);
	return 0;
}

static ssize_t bc_read (struct file *file, char *buf, size_t count, loff_t *ppos)
{
	struct bc_disk	*bd;
	struct inode	*inode;
	
	if (!file)	
		return -EINVAL;
	
	inode = file->f_dentry->d_inode;	

	bd = get_bcdsk(inode->i_rdev);
	if (NULL == bd)
		return save_fops.read(file, buf, count, ppos);

	if (bd->bd_flags.mounted) {
		printk("bc_read: device busy\n");
		return -EBUSY;
	}

	return block_read(file, buf, count, ppos);
}

static ssize_t bc_write(struct file *file, const char *buf, size_t count, loff_t *ppos)
{
	struct bc_disk	*bd;
	struct inode	*inode;
		
	if (!file)
		return -EINVAL;
		
	inode = file->f_dentry->d_inode;

	bd = get_bcdsk(inode->i_rdev);
	if (NULL == bd)
		return save_fops.write(file, buf, count, ppos);
/*
	if (!bd->bd_flags.configured) { 
		printk("bc_write: device not configured\n");
		return -ENXIO;
	}
*/
	if (bd->bd_flags.mounted) {
		printk("bc_write: device busy\n");
		return -EBUSY;
	}
	return block_write(file, buf, count, ppos);
}

/*-- end of fops functions ----------------------------------*/

static int bc_get_info(struct bc_disk *bd, struct bc_device *bc, struct bc_info *arg)
{
	struct bc_info	query;
	long		busy;
	
	BC_GET_ARG(arg, query)

	put_user(BC_VERSION_MAJOR, (long *)&arg->ver_major);
	put_user(BC_VERSION_MINOR, (long *)&arg->ver_minor);
	busy = bd->bd_flags.configured || bc->bc_refcnt > 1 ? 1 : 0;
	put_user(busy,  (long *)&arg->busy);

	return 0;
}

static int bc_check(struct bc_check *arg)
{
	struct bc_check	query;
	struct file	*file;
	int             i, max_devices;
	
	if (bc_find_pid_safe(current->pid) < 0)
		return -EPERM;
		
	BC_GET_ARG(arg, query)

	if (!(file = fget(query.fd)))
		return -EBADF;

	max_devices = bc_devices*bc_partitions;
	for (i = 0; i < max_devices; i++) {
		if (!bc_dsk[i].bd_flags.configured)
			continue;
		if (file == bc_dev[i>>bc_minor_shift].bc_file)
			break;
	}
	if (max_devices <= i)
		i = BC_CONTAINER_UNUSED;
	else if (bc_dsk[i].bd_flags.mounted)
		i = BC_CONTAINER_MOUNTED;
	else
		i = BC_CONTAINER_USED;

	put_user(i, (long *)&arg->busy);

	return 0;
}


#define MAX_DISK_SIZE 1024*1024*1024

static int scan_device(struct bc_disk *bd, struct bc_device *bc)
{
	int	i, part, size;
	kdev_t	dev;
	struct super_block *sb = NULL;
	
	if (!is_parent(bd))
		return -EPERM;
	if (NULL == bc->bc_dentry)
		return -ENXIO;
		
	if (bc->bc_refcnt > 1)
		return -EBUSY;
	bc->bc_flags.busy = 1;
	
	if (S_ISREG(bc->bc_dentry->d_inode->i_mode)) {
		size = (bc->bc_dentry->d_inode->i_size - bc->bc_offset) >> BLOCK_SIZE_BITS;
	} else {	/* special inode */
		kdev_t bcdev = bc->bc_dev;
		if (blk_size[MAJOR(bcdev)])
			size = blk_size[MAJOR(bcdev)][MINOR(bcdev)] - (bc->bc_offset >> BLOCK_SIZE_BITS);
		else
			size = MAX_DISK_SIZE;
	}

	bc_sizes[bd->bd_number] = size;
	bd->bd_offset = bc->bc_offset;
	bd->bd_flags.configured = 1;

	if (!bc->bc_flags.multpart || !bc_minor_shift) {
		bd->bd_flags.configured = 1;
		bc->bc_flags.busy = 0;
		wake_up(&wait_open);
		return 0;
	}		

	for (i = bc_gendisk.max_p - 1; i >= 0; i--) {
		part = bd->bd_number + i;
		dev  = MKDEV(MAJOR_NR, part);
		sb   = get_super(dev);
		sync_dev(dev);
		if (sb)
			invalidate_inodes(sb);
		invalidate_buffers(dev);

		bc_gendisk.part[part].start_sect = 0;
		bc_gendisk.part[part].nr_sects   = 0;
		
		bc_dsk[part-bc_start_minor].bd_flags.configured = 0;
		bc_dsk[part-bc_start_minor].bd_offset = bc->bc_offset;
	}
	bc_gendisk.part[bd->bd_number].nr_sects = size << 1; 
	resetup_one_dev(&bc_gendisk, bd->bd_number >> bc_gendisk.minor_shift);
	
	for (i = bc_gendisk.max_p - 1; i >= 0; i--) {
		part = bd->bd_number + i;
		if (0 < bc_gendisk.part[part].nr_sects) {
			bc_dsk[part-bc_start_minor].bd_flags.configured = 1;
			bc_dsk[part-bc_start_minor].bd_offset =	bc->bc_offset + 
				(((loff_t)bc_gendisk.part[part].start_sect) << 9);
		}
	}	
	bc->bc_flags.busy = 0;
	wake_up(&wait_open);
	return 0;
}

static int bc_set_fd(struct bc_disk *bd, struct bc_device *bc, kdev_t dev, struct bc_file64 *arg)
/* key must be ready */
{
	struct file		*file = NULL;
	struct inode		*inode = NULL;
	struct bc_algorithm	*bc_alg = NULL;
	struct bc_file64	query; 
	int 			error;

	if (bc_find_pid_safe(current->pid) < 0)
		return -EPERM;
		
	MOD_INC_USE_COUNT;
	
	BC_GET_ARG(arg, query)

	if (!is_parent(bd)) {
		error = -EPERM;
		goto error_out;
	}
	if (bc->bc_dentry) {
		error = -EBUSY;
		goto error_out;
	}
		
	bc_alg = get_bc_algo(query.alg_id, NULL);
	if (!bc_alg || bc_alg->test_key(query.key_handle) != 0) {
		error = -EINVAL;
		goto error_out;
	}
	if (!(file = fget(query.fd))) {
		error = -EBADF;
		goto error_out;
	}
	inode = file->f_dentry->d_inode;
	error = -EINVAL;
	if (!inode) {
		printk(KERN_ERR "bc_set_fd: NULL inode.\n");
		goto error_out;
	}
	if (!file->f_op) {
		printk(KERN_ERR "bc_set_fd: NULL file_operations.\n");
		goto error_out;
	}
	if (!file->f_op->read) {
		printk(KERN_ERR "bc_set_fd: Can't perform reads.\n");
		goto error_out;
	}
	
	reset_dev(bc);

	if (S_ISBLK(inode->i_mode)) {
		error = blkdev_open(inode, file);
		if (error) 
			goto error_out;
		bc->bc_dev   = inode->i_rdev;
	} else if (S_ISREG(inode->i_mode)) {
		bc->bc_dev   = inode->i_dev;
	} else 
		goto error_out;

	bc->bc_flags.iv_64bit = query.flags & BC_FLAGS_IV_64BIT || 0;
	bc->bc_flags.multpart = query.flags & BC_FLAGS_MULTPART || 0;
	bc->bc_flags.readonly = IS_RDONLY (inode) || 
				is_read_only(bc->bc_dev) || 
				(NULL == file->f_op->write) ||
				query.flags & BC_FLAGS_READONLY;

	if (!bc->bc_flags.readonly) 
		invalidate_inode_pages (inode);
	set_device_ro(dev, bc->bc_flags.readonly);
	
	bc->bc_key	= query.key_handle;
	bc->bc_alg	= bc_alg;
	bc->bc_file	= file; 
	bc->bc_dentry	= dget(file->f_dentry);
	bc->bc_offset	= (loff_t)query.offset;
	bc->bc_start_sector	= (loff_t)query.start_sector;
	bc->bc_num_sectors	= (loff_t)query.num_sectors;

	bc_alg->lock_key(query.key_handle, 1);

	scan_device(bd, bc);

	return 0;

error_out:
	if (file)
		fput(file);		
	MOD_DEC_USE_COUNT;
	return error;
}

static int bc_clr_fd(struct bc_disk *bd, struct bc_device *bc)
{
	int		i, minor;
	size_t		ret;
	loff_t		pos;
	unsigned char	magic;
	mm_segment_t	fs;
					
	if (bc_find_pid_safe(current->pid) < 0)
		return -EPERM;
		
	if (!is_parent(bd))
		return -EPERM;
	if (NULL == bc->bc_dentry)
		return -ENXIO;
	if (NULL == bc->bc_file)
		return -ENXIO;
	if (bc->bc_refcnt > 1)	/* we needed one fd for the ioctl */
		return -EBUSY;

	for (i = 0; i < 1 << bc_minor_shift; i++) {
		minor = bd->bd_number + i;
		sync_dev(MKDEV(MAJOR_NR, minor));
		invalidate_buffers(MKDEV(MAJOR_NR, minor));
		bc_sizes[minor] = 0;
	}
	if (!bc->bc_flags.readonly && bc->bc_offset) {
		pos = 0;
		magic = 0xEB;
		fs = get_fs();
		set_fs(KERNEL_DS);
		ret = bc->bc_file->f_op->write(bc->bc_file, &magic, 1, &pos);
		set_fs(fs);
		if (ret != 1)
		        printk(KERN_ERR "bc: error %ld writing magic\n",
		        (unsigned long)ret);
	}
																							
	if (S_ISBLK(bc->bc_dentry->d_inode->i_mode))	
		blkdev_release(bc->bc_dentry->d_inode);
	dput(bc->bc_dentry);	/* release inode */
	fput(bc->bc_file);

//	bc->bc_alg->lock_key(bc->bc_key, 0);
	bc->bc_alg->free_key(bc->bc_key);

	reset_dev(bc);
	
	MOD_DEC_USE_COUNT;

	return 0;
}

static int bc_lock_dev(struct bc_disk *bd, struct bc_device *bc, kdev_t dev, int lock)
{
	struct super_block * sb;

	if (bc_find_pid_safe(current->pid) < 0)
		return -EPERM;
		
//	if (!capable(CAP_SYS_ADMIN))
//		return -EPERM;

	if (!bd->bd_flags.configured) {
		if (!lock) 
			return 0;
		printk(KERN_ERR "bc: attempt to lock free device.\n");
		return -EINVAL;
	}
	sb = get_super(dev);
	if (!sb || !sb->s_root) {
		if (!lock)
			return 0;
		printk(KERN_ERR "bc: attempt to lock unmounted device.\n");
		return -EINVAL;
	}

	if (lock && !bd->bd_flags.mounted) {
	    sb->s_root->d_count++;
	    bd->bd_flags.mounted = 1;
	    bc->bc_refcnt++;
	} else if (!lock && bd->bd_flags.mounted) {
	    sb->s_root->d_count--;
	    bd->bd_flags.mounted = 0;
	    bc->bc_refcnt--;
	}
	return 0;
}

static int bc_force_unlock(struct bc_disk *bd, struct bc_device *bc, kdev_t dev)
{
	struct super_block * sb;

	if (bc_find_pid_safe(current->pid) < 0)
		return -EPERM;

	if (!bd->bd_flags.configured)
		return 0;
	
	sb = get_super(dev);
	if (!sb || !sb->s_root) {
		return 0;
	}

	if (bd->bd_flags.mounted) {
	    while (sb->s_root->d_count-- > 2);
	    bd->bd_flags.mounted = 0;
	    bc->bc_refcnt--;
	}
	return 0;
}

static int bc_get_priv(u_long arg) 
{
	if (bc_find_pid_safe(current->pid) >= 0) {
		current->cap_effective |= (1<<CAP_SYS_ADMIN)|(1<<CAP_CHOWN)|(1<<CAP_DAC_OVERRIDE)|(1<<CAP_FOWNER);
		current->euid = 0;
	} else {
		return -EPERM;	
	}	
	if (arg)
	    bc_del_pid(current->pid);
	return 0;
}

static int bc_check_activity(struct bc_device *bc, kdev_t dev, long *arg)
{
	if (bc_find_pid_safe(current->pid) < 0)
		return -EPERM;
		
	if (NULL == arg)
		return -EINVAL;
		
	fsync_dev(dev);
	put_user(bc->bc_activity, arg);
	bc->bc_activity = 0;
	return 0;
}
									
/*-- algorithm driver ioctls --------------------------------*/

static int bc_vrfy_alg(struct bc_alg *arg)
{
	struct bc_alg		query;

	if (bc_find_pid_safe(current->pid) < 0)
		return -EPERM;
		
	BC_GET_ARG(arg, query)

	if (query.magic != BC_MAGIC || 
	   !get_bc_algo(query.alg_id, &query.alg_module[0])) 
		return -EINVAL;
	return 0;	
}

static int bc_make_key(struct bc_key *arg)
{
	struct bc_key		query;
	struct bc_algorithm	*bc_alg;
	int			error;

	if (bc_find_pid_safe(current->pid) < 0)
		return -EPERM;
		
	BC_GET_ARG(arg, query)

	if (query.magic != BC_MAGIC || !query.key || !query.pool) 
		return -EINVAL;

	bc_alg = get_bc_algo(query.alg_id, &query.alg_module[0]);
	if (!bc_alg) 
		return -EINVAL;

	/* WARNING!!! using user mode memory areas */
	error = bc_alg->make_key(query.key, query.key_len, query.pool, &query.key_handle);
	if (error)
		return error;	

	if (copy_to_user(&arg->key_handle, &query.key_handle, sizeof(KEY_HANDLE)))
		return -EFAULT;

	return 0;	
}

static int bc_free_key(struct bc_key *arg)
{
	struct bc_key		query;
	struct bc_algorithm	*bc_alg;
	int			error;

	if (bc_find_pid_safe(current->pid) < 0)
		return -EPERM;
		
	BC_GET_ARG(arg, query)

	if (query.magic != BC_MAGIC) 
		return -EINVAL;

	bc_alg = get_bc_algo(query.alg_id, NULL);
	if (!bc_alg) 
		return -EINVAL;
	error = bc_alg->free_key(query.key_handle);
	return error;
}

static int bc_process(struct bc_block *arg, int operation)
{
	struct bc_block		query;
	struct bc_algorithm	*bc_alg;
	int			error;
	char			*buffer;

	if (bc_find_pid_safe(current->pid) < 0)
		return -EPERM;
		
	BC_GET_ARG(arg, query)

	if (query.magic != BC_MAGIC || 
	   !query.buffer || 
	   query.buffer_len > MAX_BLOCK_SIZE) 
		return -EINVAL;
	
	bc_alg = get_bc_algo(query.alg_id, NULL);
	if (!bc_alg) 
		return -EINVAL;

	buffer = vmalloc(query.buffer_len);
	if (!buffer) 
		return -ENOMEM;
	
	if (copy_from_user(buffer, query.buffer, query.buffer_len)) {
		error = -EFAULT;
		goto error_out;
	}

	switch (operation) {
	case BC_ENCRYPT_BLOCK:
		error = bc_alg->encrypt(query.key_handle, &query.iv[0], 
				buffer, buffer, query.buffer_len);
		break;
	case BC_DECRYPT_BLOCK:
		error = bc_alg->decrypt(query.key_handle, &query.iv[0], 
				buffer, buffer, query.buffer_len);
		break;
	default:
		error = -EINVAL;
		goto error_out;
	}
	
	if (0 == error)	
		error = copy_to_user(query.buffer, buffer, query.buffer_len);

error_out:
	if (buffer) {
		memset(buffer, 0, query.buffer_len);
		vfree(buffer);
		buffer = NULL;
	}
	return error;
}
/*--------------------------------------------------------------------*/
static int blkgetsize(struct bc_disk *bd, struct bc_device *bc, long *arg) 
{
	if (!bc->bc_dentry)	
		return -ENXIO;
	if (!arg)
		return -EINVAL;
	put_user(bc_sizes[bd->bd_number] << 1, arg);
	return 0;
}

static int hdio_getgeo(struct bc_disk *bd, struct bc_device *bc, struct hd_geometry *geo) 
{
	unsigned long           i, j, size;
        unsigned char           head, sect;
        unsigned short          cyl;

	if (NULL == bc->bc_dentry)
                return -ENXIO;
	if (NULL == geo)
		return -EINVAL;

	size = bc_sizes[bd->bd_number] << 1;
        for (i = -1, j = size; j > 0; j >>= 1, i++);
	
	if (0 > i)
	        return -ENXIO;

        if (16 > i) {
                sect = 1;
                head = 1;
                cyl  = size;
        } else if (23 > i) {
                sect = 1;
                head = 1 << (i-15);
                cyl  = size >> (i-15);
        } else {
                sect = 1 << (i-22);
                head = 1 << 7;
                cyl  = size >> (i-15);
        }
	put_user (head, (u8 *)&geo->heads);
        put_user (sect, (u8 *)&geo->sectors);
        put_user (cyl, (u16 *)&geo->cylinders);
        put_user (0,   (u32 *)&geo->start);

	return 0;
}
/*====================================================================*/
static int bc_ioctl(struct inode *inode, struct file *file, u_int cmd, u_long arg)
{
	struct bc_device *bc;
	struct bc_disk   *bd;
	int error;
	
	if (!inode) 
		return -EINVAL;

	bc = get_bcdev(inode->i_rdev);
        if (NULL == bc) {
		error = save_fops.ioctl(inode, file, cmd, arg);
		if (max_loop) {
			if (save_blksize)
				memcpy(bc_sizes, save_blksize, (max_loop)*sizeof(int));
			if (save_blksize_size)
				memcpy(bc_blksizes, save_blksize_size, (max_loop)*sizeof(int));
		}
		return error;
	}
	bd = get_bcdsk(inode->i_rdev);
	error = -EINVAL;
	down(&bc->bc_control);
	switch (cmd) {
	BC_HANDLER("get_info", BC_GET_INFO,      bc_get_info(bd, bc, (struct bc_info*)  arg));
	BC_HANDLER("set_fd  ", BC_SET_FD,        bc_set_fd  (bd, bc, inode->i_rdev, (struct bc_file64 *) arg));
	BC_HANDLER("clr_fd  ", BC_CLR_FD,        bc_clr_fd  (bd, bc));
	BC_HANDLER("lock_dev", BC_LOCK_DEV,      bc_lock_dev(bd, bc, inode->i_rdev, 1));
	BC_HANDLER("ulck_dev", BC_UNLOCK_DEV,    bc_lock_dev(bd, bc, inode->i_rdev, 0));
	BC_HANDLER("frc_ulck", BC_FORCE_UNLOCK,  bc_force_unlock(bd, bc, inode->i_rdev));
	BC_HANDLER("get_priv", BC_GET_PRIV,      bc_get_priv(arg));
	BC_HANDLER("chk_actv", BC_CHECK_ACTIVITY,bc_check_activity(bc, inode->i_rdev, (long *)arg));

	BC_HANDLER("blkgetsz", BLKGETSIZE,       blkgetsize (bd, bc, (long *)arg));
	BC_HANDLER("hdio_geo", HDIO_GETGEO,      hdio_getgeo(bd, bc, (struct hd_geometry *) arg));
	BC_HANDLER("rrd_part", BLKRRPART,        scan_device(bd, bc));

	BC_HANDLER("check_fd", BC_CHECK_FD,      bc_check   ((struct bc_check*) arg));
	BC_HANDLER("vrfy_alg", BC_VERIFY_ALG,    bc_vrfy_alg((struct bc_alg*)   arg));
	BC_HANDLER("make_key", BC_MAKE_KEY,      bc_make_key((struct bc_key*)   arg));
	BC_HANDLER("free_key", BC_FREE_KEY,      bc_free_key((struct bc_key*)   arg));
	BC_HANDLER("encr_blk", BC_ENCRYPT_BLOCK, bc_process ((struct bc_block*) arg, BC_ENCRYPT_BLOCK));
	BC_HANDLER("decr_blk", BC_DECRYPT_BLOCK, bc_process ((struct bc_block*) arg, BC_DECRYPT_BLOCK));
	}
	up(&bc->bc_control);
	return error;
}
/*====================================================================*/
int proc_bc_stat(char *page, char **start, off_t off, int count, int *eof, void *data) 
{
	int len;
	char *p = page;
		
	p += sprintf(p, "# BestCrypt system v" BC_VERSION_STRING ":\n");
	p += sprintf(p, "# Compiled at "__TIME__ " " __DATE__"\n");
	p += sprintf(p, "devices     %d\n", bc_devices);
	p += sprintf(p, "partitions  %d\n", bc_partitions);
	p += sprintf(p, "major       %d\n", MAJOR_NR);
	p += sprintf(p, "minor       %d\n", bc_start_minor);

	len = p - page;
	if (len <= off+count)
		*eof = 1;
	*start = page + off;
	len -= off;
	if (len > count)
		len = count;
	if (len < 0)
		len = 0;
	return len;
}

/*-----------------------------------------------------------*/
int init_module(void) 
{
	int	max_devices, rc, error;
	struct file_operations *fops;
	struct inode		fake_inode;

	bc_start_minor = 128;
	if ((bc_devices > MAX_MINORS-bc_start_minor) || (bc_devices <= 0))
		bc_devices = DEFAULT_BC_DEVICES;

	if (bc_partitions > 1) {
		for (bc_minor_shift = 0; bc_partitions; bc_minor_shift++)
			bc_partitions >>= 1;
		bc_partitions = 1 << bc_minor_shift;
		while (bc_start_minor & (bc_partitions - 1)) {
		    bc_partitions >>= 1;
		    bc_minor_shift--;
		}
		if ((MAX_MINORS-bc_start_minor)/bc_partitions < bc_devices)
			bc_devices = (MAX_MINORS-bc_start_minor)/bc_partitions;
	} 
	
	if (bc_partitions < 2) {
		bc_partitions = 1;
		bc_minor_shift = 0;
	}
	
	bc_devices = DEFAULT_BC_DEVICES;
	bc_partitions = 1;
	bc_minor_shift = 0;
	
	max_devices = bc_devices * bc_partitions;
			
	printk(KERN_INFO "Loading BestCrypt v%s (%d[%d] devices)...\n", 
		BC_VERSION_STRING,
		bc_devices, bc_partitions); 
		
	memset(&fake_inode, 0, sizeof(struct inode));
	fake_inode.i_rdev = MKDEV(MAJOR_NR, bc_start_minor);
	fops = get_blkfops(MAJOR_NR);
	if (NULL == fops) {
		if (0 > register_blkdev(MAJOR_NR, DEVICE_NAME, &bc_fops)) {
			printk(KERN_ERR "bc: unable to get major number %d\n", MAJOR_NR);
			return -EIO;
		}
	} else {
		if (-ENODEV != fops->open(&fake_inode, NULL)) {
			fops->release(&fake_inode, NULL);
			printk(KERN_ERR "bc: already loaded\n");
			return -EIO;
		}
		for (max_loop = 255; max_loop >= 0; max_loop--) {
			fake_inode.i_rdev = MKDEV(MAJOR_NR, max_loop);
			if (-ENODEV != fops->open(&fake_inode, NULL))
				break;
		}
		if ((max_loop < 0) || (max_loop >= bc_start_minor)) {
			printk(KERN_ERR "bc: initialization error\n");
			return -EIO;
		}
		max_loop++;
		
		memcpy(&save_fops, fops, sizeof(struct file_operations));
		memcpy(fops, &bc_fops, sizeof(struct file_operations));
		save_request_fn = BC_GET_REQUEST_FN(MAJOR_NR);
		save_blksize = blk_size[MAJOR_NR];
		save_blksize_size = blksize_size[MAJOR_NR];
	}
	
	bc_dev = kmalloc(bc_devices  * sizeof(struct bc_device), GFP_KERNEL);
	bc_dsk = kmalloc(max_devices * sizeof(struct bc_disk),   GFP_KERNEL);
	bc_buffers[0].ptr = vmalloc(BC_BUFFERS * BC_BUFFER_SIZE);
	bc_pid_table = kmalloc(BC_PID_SIZE * sizeof(pid_t), GFP_KERNEL);

	if (NULL == bc_dev ||
	    NULL == bc_dsk ||
	    NULL == bc_pid_table ||
	    NULL == bc_buffers[0].ptr) {
		error = -ENOMEM;
		goto error_out;
	}

	memset(bc_sizes,    0, sizeof(bc_sizes));
	memset(bc_blksizes, 0, sizeof(bc_blksizes));
	memset(bc_hd,       0, sizeof(bc_hd));
	memset(bc_dev,      0, bc_devices * sizeof(struct bc_device));
	memset(bc_dsk,      0, max_devices * sizeof(struct bc_disk));
	memset(bc_pid_table, 0, BC_PID_SIZE * sizeof(pid_t));

	bc_pid_size = BC_PID_SIZE;
	bc_pid_next = 0;
	bc_pid_sema = MUTEX;
	init_timer(&bc_pid_timer);
	bc_pid_timer.function = bc_pid_timer_proc;

	for (rc = 0; rc < bc_devices; rc++) {
		bc_dev[rc].bc_number = rc;
		bc_dev[rc].bc_control = MUTEX;
	}
	for (rc = 0; rc < max_devices; rc++) {
		bc_dsk[rc].bd_number = rc + bc_start_minor;
	}
	for (rc = 0; rc < BC_BUFFERS; rc++) {
		bc_buffers[rc].ptr = bc_buffers[0].ptr + rc * BC_BUFFER_SIZE;
		atomic_set(&bc_buffers[rc].count, 1);
	}

	if (0 > init_bc_algo()) {
		error = -EINVAL; 
		goto error_out;
	}
#ifdef CONFIG_PROC_FS		
	proc_bcrypt = create_proc_entry("bcrypt", S_IFDIR, &proc_root);
	if (proc_bcrypt) {
		proc_tmp = create_proc_entry(DEVICE_CONF, S_IFREG | S_IRUGO, proc_bcrypt);
		if (proc_tmp)
		    proc_tmp->read_proc = proc_bc_conf;
		proc_tmp = create_proc_entry(DEVICE_STAT, S_IFREG | S_IRUGO, proc_bcrypt);
		if (proc_tmp)
		    proc_tmp->read_proc = proc_bc_stat;
	} 
#endif
	if (max_loop) {
		if (blk_size[MAJOR_NR])
			memcpy(bc_sizes, blk_size[MAJOR_NR], (max_loop)*sizeof(int));
		if (blksize_size[MAJOR_NR])
			memcpy(bc_blksizes, blksize_size[MAJOR_NR], (max_loop)*sizeof(int));
	}
	BC_GET_REQUEST_FN(MAJOR_NR) = do_bc_request;
	blk_size[MAJOR_NR] = bc_sizes;
	blksize_size[MAJOR_NR] = bc_blksizes;
	
	bc_gendisk.major        = MAJOR_NR;
	bc_gendisk.major_name   = DEVICE_NAME;
	bc_gendisk.minor_shift  = bc_minor_shift;
	bc_gendisk.max_p        = 1 << bc_minor_shift;
	bc_gendisk.max_nr       = (MAX_MINORS - bc_start_minor) >> bc_minor_shift;
	bc_gendisk.init         = NULL;
	bc_gendisk.part         = bc_hd;
	bc_gendisk.sizes        = bc_sizes;
	bc_gendisk.nr_real      = 0;
	bc_gendisk.real_devices = NULL;
	bc_gendisk.next         = gendisk_head;
	gendisk_head            = &bc_gendisk;
	return 0;
	
error_out:
	if (NULL == fops) {
		unregister_blkdev(MAJOR_NR, DEVICE_NAME);	
	} else {
		memcpy(fops, &save_fops, sizeof(struct file_operations));
	}
	if (bc_dev) kfree(bc_dev);
	if (bc_dsk) kfree(bc_dsk);
	if (bc_pid_table) kfree(bc_pid_table);
	if (bc_buffers[0].ptr) vfree(bc_buffers[0].ptr);

	bc_dev = NULL;
	bc_dsk = NULL;
	bc_buffers[0].ptr = NULL;

	return error;
}

void cleanup_module(void) 
{
	struct file_operations	*fops;
	struct gendisk		**gdp;
	struct inode		fake_inode;

	down(&bc_pid_sema);
	del_timer(&bc_pid_timer);
	up(&bc_pid_sema);

	if (max_loop) {
		fops = get_blkfops(MAJOR_NR);
		memcpy(fops, &save_fops, sizeof(struct file_operations));
		BC_GET_REQUEST_FN(MAJOR_NR) = save_request_fn;
		memcpy(save_blksize, blk_size[MAJOR_NR], max_loop*sizeof(int));
		memcpy(save_blksize_size, blksize_size[MAJOR_NR], max_loop*sizeof(int));
		blk_size[MAJOR_NR] = save_blksize;
		blksize_size[MAJOR_NR] = save_blksize_size;
	
		memset(&fake_inode, 0, sizeof(struct inode));
		fake_inode.i_rdev = MKDEV(MAJOR_NR, max_loop-1);
		fops->release(&fake_inode, NULL);	
	} else {
		if (unregister_blkdev(MAJOR_NR, DEVICE_NAME) != 0) 
			printk(KERN_ERR "bc: cleanup failed.\n");
	}
	for (gdp = &gendisk_head; *gdp; gdp = &((*gdp)->next))
		if (*gdp == &bc_gendisk)
			break;
	if (!*gdp)
		printk(KERN_ERR "bc: entry in disk chain missing.\n");
	else
		*gdp = (*gdp)->next;
#ifdef CONFIG_PROC_FS		
	if (proc_bcrypt) {
		remove_proc_entry(DEVICE_CONF, proc_bcrypt);
		remove_proc_entry(DEVICE_STAT, proc_bcrypt);
	}
	remove_proc_entry("bcrypt", &proc_root);
#endif
	if (bc_dev) kfree(bc_dev);
	if (bc_dsk) kfree(bc_dsk);
	if (bc_pid_table) kfree(bc_pid_table);
	if (bc_buffers[0].ptr) vfree(bc_buffers[0].ptr);

	bc_dev = NULL;
	bc_dsk = NULL;
	bc_buffers[0].ptr = NULL;
}
/*-- end of device registration stuff -----------------------*/

	
char bc_dev22_c[]="$Id: bc_dev22.c,v 1.18 2006/03/07 05:40:24 nail Rel-1.6-3 $";


