/*******************************************************************
 *	Copyright (c) 1994-1998 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	File:		driver/bc_dev.c
 *
 *	Description:	BestCrypt pseudo-device driver.
 *
 *	Scope:		BestCrypt pseudo-device driver
 *
 *	Platforms:	Linux
 *
 *	Ideas:		Igor M. Arsenin
 *			Serge A. Frolov
 *			Vitaliy G. Zolotarev
 *
 *	Author:		Nail R. Kaipov
 *
 *	Created:	10-Nov-1998
 *
 *	Revision:	06-Aug-1999.	new strategy routine
 *			15-Sep-1999.	2.3.x kernels support
 *			11-Oct-1999.	strategy bugfix
 *			15-Dec-1999.	driver refinement
 *			21-Mar-2000.	bugfix for large blocks (v0.4b-2)
 *			15-Jun-2000.	partition support
 *			15-Nov-2000.	major rewrite. 2.4.0 version
 *			01-Feb-2001.	minor fixes. partitions cleanup.
 *			02-May-2001.	SMP bugfix, /proc update
 *			06-Jun-2001.	writing magic in bc_clr_fd
 *			10-Oct-2001.	bc_lock_dev() fixed
 *			04-Nov-2003.	64-bit IV added.
 *
 *******************************************************************/

#ifndef EXPORT_SYMTAB
#define EXPORT_SYMTAB
#endif

#include "bc_cfg.h"
#include <linux/module.h>
#include <linux/fs.h>			/* inode et al            */
#include <linux/proc_fs.h>		/* proc_fs support        */
#include <linux/devfs_fs_kernel.h>	/* devfs_fs support       */
#include <linux/file.h>			/* fget()/fput() pair     */
#include <linux/vmalloc.h>		/* vmalloc()              */
#include <asm/uaccess.h>		/* put_user()             */
#include <asm/atomic.h>			/* atomic operations      */
#include <linux/hdreg.h>		/* hd_geometry            */
#include "bc_types.h"			/* bc types               */
#include "bc_ioctl.h"			/*           ioctl codes  */
#include "bc_mgr.h"
#include "bc_dev.h"			/* bc types        codes  */

#define DEVICE_OFF(dev)
#include <linux/blk.h>

#define __KERNEL_SYSCALLS__
#include <linux/unistd.h>

#ifndef GFP_NOIO
#define GFP_NOIO GFP_BUFFER
#endif

static int		bc_devices = DEFAULT_BC_DEVICES;
static int		bc_partitions = DEFAULT_BC_PARTITIONS;
static int		bc_minor_shift = 0;

struct bc_device	*bc_dev;		/* bc devices' array  */
static struct bc_disk   *bc_dsk;		/* bc disks array     */
static int		*bc_sizes;		/**/
static int		*bc_blksizes;		/**/
static struct hd_struct	*bc_hd;			/**/

/* pid control table support */
#define BC_PID_SIZE 64
static pid_t            	*bc_pid_table;
static int			bc_pid_size;
static int			bc_pid_next;
static struct semaphore 	bc_pid_sema;
static struct timer_list	bc_pid_timer;

static devfs_handle_t		bc_hdevfs;
static struct gendisk		bc_gendisk;
static struct proc_dir_entry    *proc_bcrypt;
static DECLARE_WAIT_QUEUE_HEAD(wait_open);
static DECLARE_WAIT_QUEUE_HEAD(wait_bc);

EXPORT_SYMBOL(register_bc_algo);
EXPORT_SYMBOL(unregister_bc_algo);

MODULE_PARM(bc_devices, "i");
MODULE_PARM_DESC(bc_devices, "Max. number of devices (1-256).");

MODULE_PARM(bc_partitions, "i");
MODULE_PARM_DESC(bc_partitions, "Max. partitions per device (0-63).");

MODULE_AUTHOR("Jetico, Inc.");
MODULE_DESCRIPTION("BestCrypt encrypted block device driver.");
#ifdef MODULE_LICENSE
MODULE_LICENSE("Jetico, Inc.");
#endif

static int     bc_ioctl     (struct inode *,  struct file *, unsigned int, unsigned long);
static int     bc_open      (struct inode *,  struct file *);
static int     bc_release   (struct inode *,  struct file *);
/*
static struct file_operations bc_fops;
*/
static struct block_device_operations bc_bdops = {
	ioctl  :bc_ioctl,
	open   :bc_open,
	release:bc_release
};

/*-----------------------------------------------------------*/
static inline struct bc_device *get_bcdev(kdev_t dev)
{
	int 	mn = MINOR(dev);
	if (MAJOR_NR == MAJOR(dev) && MAX_MINORS > MINOR(dev)) {
		mn >>= bc_minor_shift; 	
		if (0 <= mn && bc_devices > mn)
			return bc_dev+mn;
	}
	return NULL; /* -ENODEV error */
}

static inline struct bc_disk *get_bcdsk(kdev_t dev) 
{
	int	mn = MINOR(dev);
	if (MAJOR_NR == MAJOR(dev) && MAX_MINORS > MINOR(dev)) {
		if (0 <= mn && bc_devices * bc_partitions > mn)
			return bc_dsk+mn;
	}
	return NULL;	
}

static inline int is_parent(struct bc_disk *bd) 
{
	if (NULL != bd)
		return !(bd->bd_number & ((1 << bc_minor_shift) - 1));
	return 0;
}

static inline void reset_dev(struct bc_device *bc)
{	
	int             i, j;

	if (NULL == bc)
		return;

	bc->bc_offset = 0L;
	bc->bc_start_sector = 0L;
	bc->bc_num_sectors  = 0L;
	bc->bc_dev    = 0;
	bc->bc_key    = 0;
	bc->bc_alg    = NULL;
	bc->bc_file   = NULL;
	bc->bc_dentry = NULL;
	atomic_set(&bc->bc_pending, 0);
	spin_lock_init(&bc->bc_lock);
	init_MUTEX_LOCKED(&bc->bc_thread);
	init_MUTEX_LOCKED(&bc->bc_run);
	bc->bc_bh     = NULL;
	bc->bc_bhtail = NULL;
	bc->bc_gfpmask = 0;
	bc->bc_save_fops = NULL;
	bc->bc_activity = 0;
	memset(&bc->bc_flags, 0, sizeof(bc->bc_flags));
	if (NULL != bc->bc_buffer)
		vfree(bc->bc_buffer);
	bc->bc_buffer = NULL;

	i = bc->bc_number << bc_minor_shift;
	j = i + (1 << bc_minor_shift);
	memset(&bc_dsk[i], 0, (j-i) * sizeof(struct bc_disk));
	for (; i < j; i++)
		bc_dsk[i].bd_number = i;
}
/*-----------------------------------------------------------*/
static inline int bc_find_pid(pid_t pid) 
{
	register int i;
	for (i = 0; i < bc_pid_next; i++)
		if (pid == bc_pid_table[i])
			return i;
	return -1;
}

static int bc_find_pid_safe(pid_t pid)
{	
	int i;
	down(&bc_pid_sema);
	i = bc_find_pid(pid);
	up(&bc_pid_sema);
	return i;
}

static int bc_add_pid(pid_t pid) 
{
	pid_t	*tmp;
	
	down(&bc_pid_sema);
	if (bc_find_pid(pid) > -1) {
		up(&bc_pid_sema);
		return 0;
	}
	if (bc_pid_next < bc_pid_size) {
		bc_pid_table[bc_pid_next] = pid;
		bc_pid_next++;
	} else {
		tmp = kmalloc((bc_pid_size+BC_PID_SIZE) * sizeof(pid_t), GFP_KERNEL);
		if (NULL == tmp) {
			up(&bc_pid_sema);
			return -ENOMEM;
		}
		memcpy(bc_pid_table, tmp, bc_pid_size * sizeof(pid_t));
		kfree(bc_pid_table);
		bc_pid_table = tmp;
		bc_pid_size += BC_PID_SIZE;
		bc_pid_table[bc_pid_next] = pid;
		bc_pid_next++;
	}
	if (1 == bc_pid_next) {
		bc_pid_timer.expires = jiffies + 5 * HZ;
		add_timer(&bc_pid_timer);
	}
	up(&bc_pid_sema);
	return 0;
}

static int bc_del_pid(pid_t pid) 
{
	int	i;
	pid_t	*tmp;

	down(&bc_pid_sema);

	i = bc_find_pid(pid);
	if (0 > i) {
		up(&bc_pid_sema);
		return 0;
	}
	bc_pid_next--;
	if (bc_pid_next)
	        bc_pid_table[i] = bc_pid_table[bc_pid_next];
	bc_pid_table[bc_pid_next] = 0;
		
	if (bc_pid_size-bc_pid_next > 2*BC_PID_SIZE) {
		tmp = kmalloc((bc_pid_size-BC_PID_SIZE) * sizeof(pid_t), GFP_KERNEL);
		if (NULL == tmp) {
			up(&bc_pid_sema);
			return 0;
		}
		bc_pid_size -= BC_PID_SIZE;
		memcpy(bc_pid_table, tmp, bc_pid_size * sizeof(pid_t));
		kfree(bc_pid_table);
		bc_pid_table = tmp;
	}
	up(&bc_pid_sema);
	return 0;
}

static void bc_pid_timer_proc(unsigned long unused) 
{
	register int i;

	down(&bc_pid_sema);
#ifdef CONFIG_SMP
	write_lock_irq(&tasklist_lock);
#endif	
	for (i = 0; i < bc_pid_next; i++) {
		if (NULL == find_task_by_pid(bc_pid_table[i])) {
			bc_pid_next--;
			if (bc_pid_next)
				bc_pid_table[i] = bc_pid_table[bc_pid_next];
			bc_pid_table[bc_pid_next] = 0;
			i--;
		}
	}
#ifdef CONFIG_SMP
        write_unlock_irq(&tasklist_lock);
#endif
	
	if (bc_pid_next) {
		bc_pid_timer.expires = jiffies + 5 * HZ;
		add_timer(&bc_pid_timer);
	}
	up(&bc_pid_sema);
}

/*-----------------------------------------------------------*/
static int bc_thread(void *data)
{
	struct bc_device   *bc = (struct bc_device *)data;
	struct buffer_head *bh;
	struct bc_disk     *bd;
	struct file	   *file;
	size_t		   len, ret, size;
	loff_t		   pos, sec;
	char		   *dst, *buf = NULL;
	u16		   iv[4];
	u64		   iv64, iv64le;
	mm_segment_t	   fs;
	int		   i, uptodate, rw;

	daemonize();
	sprintf(current->comm, "bcrypt%d", bc->bc_number);

#ifdef SCHED_OTHER
	spin_lock_irq(&current->sigmask_lock);
	sigfillset(&current->blocked);
	flush_signals(current);
	spin_unlock_irq(&current->sigmask_lock);
	current->policy = SCHED_OTHER;
#else
	spin_lock_irq(&current->sighand->siglock);
	sigfillset(&current->blocked);
	flush_signals(current);
	spin_unlock_irq(&current->sighand->siglock);
	current->policy = SCHED_NORMAL;
#endif

#ifdef _NICE_PRESENT_
	current->nice = -20;
#else
	set_user_nice(current,  -20l);
#endif
	
#ifdef PF_NOIO /* 2.4.17 kludge */
	current->flags |= PF_NOIO;
#endif 
	spin_lock_irq(&bc->bc_lock);
	bc->bc_flags.active = 1;
	atomic_inc(&bc->bc_pending);
	spin_unlock_irq(&bc->bc_lock);

	up(&bc->bc_thread);
	while (1) {
		down_interruptible(&bc->bc_run);

		i = atomic_read(&bc->bc_pending);
		if (i == 0)
			break;

		spin_lock_irq(&bc->bc_lock);
		bh = bc->bc_bh;
		if (NULL == bh) {
			spin_unlock_irq(&bc->bc_lock);
			continue;
		}
		if (bc->bc_bhtail == bh)
			bc->bc_bhtail = NULL;
		bc->bc_bh = bh->b_reqnext;
		bh->b_reqnext = NULL;
		spin_unlock_irq(&bc->bc_lock);
		uptodate = 0;
		
		bd = get_bcdsk(bh->b_rdev);
		if (!bd->bd_flags.configured && bc->bc_flags.busy) {
			bd->bd_offset = bc->bc_offset +
			(((loff_t)bc_gendisk.part[bd->bd_number].start_sect) << 9);
		}
		sec   = (loff_t)bh->b_rsector + bc->bc_start_sector;
		pos   = sec * 512 + bd->bd_offset;
		len   = bh->b_size;
		dst   = bh->b_data;
		file  = bc->bc_file;
		iv[0] = iv[1] = iv[2] = iv[3] = sec & 0xFFFF;

		iv64  = sec + 1;

		bc->bc_activity = 1;

		rw = !!test_and_clear_bit(BH_Dirty, &bh->b_state);
		if (READ == rw) {
			fs = get_fs();
			set_fs(KERNEL_DS);
			ret = file->f_op->read(file, dst, len, &pos);
			set_fs(fs);
			if (ret != len) {
				printk(KERN_ERR "bc: read error wanted %ld, read %ld\n",
					(unsigned long)len, (unsigned long)ret);
				goto error_out;
			}
			if (bc->bc_flags.iv_64bit) {
				iv64le = __cpu_to_le64(iv64);
				bc->bc_alg->decrypt(bc->bc_key, (char *)(&iv64le), dst, dst, len);
			} else {
				bc->bc_alg->decrypt(bc->bc_key, (char *)iv, dst, dst, len);
			}
		} else {
			buf = bc->bc_buffer;
			size = BC_BUFFER_SIZE;
			while (len > 0) {
				if (size > len)
					size = len;
				if (bc->bc_flags.iv_64bit) {
					iv64le = __cpu_to_le64(iv64);
					bc->bc_alg->encrypt(bc->bc_key, (char *)(&iv64le), dst, buf, size);
				} else {
					bc->bc_alg->encrypt(bc->bc_key, (char *)iv, dst, buf, size);
				}
				fs = get_fs();
				set_fs(KERNEL_DS);
				ret = file->f_op->write(file, buf, size, &pos);
				set_fs(fs);
				if (ret != size) {
					printk(KERN_ERR "bc: write error wanted %ld, wrote %ld\n",
						(unsigned long)size, (unsigned long)ret);
					goto error_out;
				}
				if (bc->bc_flags.iv_64bit) {
					iv64 += BC_BUFFER_SIZE/512;
				} else {
					iv[0] = ((int)iv[0] + BC_BUFFER_SIZE/512) & 0xFFFF;
					iv[3] = iv[2] = iv[1] = iv[0];
				}
				dst += size;
				len -= size;
			}
		}
		uptodate = 1;
error_out:
		bh->b_end_io(bh, uptodate);
		if (atomic_dec_and_test(&bc->bc_pending))
		    break;
	}
	up(&bc->bc_thread);
	return 0;
}

static int bc_make_request(request_queue_t *q, int rw, struct buffer_head *bh)
{
	int rw_ahead;
	struct bc_device *bc;
	struct bc_disk   *bd;
	
	if (!buffer_locked(bh))
		BUG();
		
	rw_ahead = 0;
	bc = get_bcdev(bh->b_rdev);
	bd = get_bcdsk(bh->b_rdev);

	if (!bc || !bd || !bc->bc_dentry || !bc->bc_dentry->d_inode) {
		printk(KERN_ERR "bc: invalid device.\n");
		goto error_out;
	}
	if (!bd->bd_flags.configured && !bc->bc_flags.busy) {
		printk(KERN_ERR "bc: device not configured yet.\n");
		goto error_out;
	}
	if (WRITE == rw) {
		if (bc->bc_flags.readonly) {
			printk(KERN_ERR "bc: attempt to write on readonly device.\n");
			goto error_out;
		}
	} else if (READA == rw) {
		rw = READ;
		rw_ahead = 1;
	} else if (READ != rw) {
		printk(KERN_ERR "bc: unknown command (%d).\n", rw);
		goto error_out;
	}

	
	spin_lock_irq(&bc->bc_lock);	
	if (!bc->bc_flags.active) {
		spin_unlock_irq(&bc->bc_lock);
		goto error_out;
	}
	atomic_inc(&bc->bc_pending);
//	spin_unlock_irq(&bc->bc_lock);	

#ifdef CONFIG_HIGHMEM
	bh = create_bounce(rw, bh);
#endif	
	if (WRITE == rw)
		set_bit(BH_Dirty, &bh->b_state);
//	spin_lock_irq(&bc->bc_lock);
	if (bc->bc_bhtail)	bc->bc_bhtail->b_reqnext = bh;
	else			bc->bc_bh = bh;
	bc->bc_bhtail = bh;
	spin_unlock_irq(&bc->bc_lock);	
	up(&bc->bc_run);
	return 0;
error_out:
	buffer_IO_error(bh);
	return 0;
}
/*-- end of strategy routine --------------------------------*/

/*-- fops functions -----------------------------------------*/
static int bc_open(struct inode *inode, struct file *file)
{
	struct bc_device *bc;

	if (NULL == inode)	
	    return -EINVAL;

	MOD_INC_USE_COUNT;

	bc = get_bcdev(inode->i_rdev);
	if (NULL == bc) {
		printk(KERN_ERR "bc_open: invalid device (%d, %d).\n",
			MAJOR(inode->i_rdev),
			MINOR(inode->i_rdev));

		MOD_DEC_USE_COUNT;

		return -ENODEV;
	}

	while (1) {
		down(&bc->bc_control);
		if (!bc->bc_flags.busy)
			break;
		up(&bc->bc_control);
		sleep_on(&wait_open);
	}
	bc->bc_refcnt++;
	up(&bc->bc_control);

	if (capable(CAP_SYS_ADMIN)) {
		bc_add_pid(current->pid);	
	}
	return 0;
}

static int bc_release(struct inode *inode, struct file *file)
{
	struct bc_device *bc;

	if (NULL == inode)
		return 0;
	bc = get_bcdev(inode->i_rdev);
	if (NULL == bc)
		return 0;
/*
	fsync() is not required here. It should
	be called before by upper level. Moreover,
	fsync() here leads to ReiserFS hangup on 
	umount. 
		
	error = fsync_dev(inode->i_rdev);
*/
	down(&bc->bc_control);
	if (0 >= bc->bc_refcnt)
		printk(KERN_WARNING "bc_release: refcount (%d) <= 0\n", bc->bc_refcnt);
	else  {
		bc->bc_refcnt--;
		MOD_DEC_USE_COUNT;
	}
	up(&bc->bc_control);
	return 0;
}

static ssize_t bc_read (struct file *file, char *buf, size_t count, loff_t *ppos)
{
	struct bc_disk		*bd;
	struct bc_device	*bc;
	struct inode		*inode;

	if (NULL == file)
		return -EINVAL;

	inode = file->f_dentry->d_inode;
	if (NULL == inode)
		return -EIO;
		
	bd = get_bcdsk(inode->i_rdev);
	if (NULL == bd)
		return -ENODEV;

	bc = get_bcdev(inode->i_rdev);
	if (NULL == bc)
		return -ENODEV;

	if (!bd->bd_flags.configured || bd->bd_flags.mounted)
		return -EIO;

	if (bc->bc_save_fops && bc->bc_save_fops->read)
		return bc->bc_save_fops->read(file, buf, count, ppos);
	else
		return -EPERM;
}

static ssize_t bc_write(struct file *file, const char *buf, size_t count, loff_t *ppos)
{
	struct bc_disk		*bd;
	struct bc_device	*bc;
	struct inode		*inode;

	if (NULL == file)
		return -EINVAL;

	inode = file->f_dentry->d_inode;
	if (NULL == inode)
		return -EIO;
		
	bd = get_bcdsk(inode->i_rdev);
	if (NULL == bd)
		return -ENODEV;

	bc = get_bcdev(inode->i_rdev);
	if (NULL == bc)
		return -ENODEV;

	if (!bd->bd_flags.configured || bd->bd_flags.mounted)
		return -EIO;

	if (bc->bc_save_fops && bc->bc_save_fops->write)
		return bc->bc_save_fops->write(file, buf, count, ppos);
	else
		return -EPERM;
}

/*-- end of fops functions ----------------------------------*/

static int bc_get_info(struct bc_disk *bd, struct bc_device *bc, struct bc_info *arg)
{
	struct bc_info	query;
	long		busy;

	BC_GET_ARG(arg, query)

	put_user(BC_VERSION_MAJOR, (long *)&arg->ver_major);
	put_user(BC_VERSION_MINOR, (long *)&arg->ver_minor);
	busy = bd->bd_flags.configured || bc->bc_refcnt > 1 ? 1 : 0;
	put_user(busy, (long *)&arg->busy);

	return 0;
}

static int bc_check(struct bc_check *arg)
{
	struct bc_check	 query;
	struct file	 *file;
	int              i, max_devices;

	BC_GET_ARG(arg, query)

	if (bc_find_pid_safe(current->pid) < 0)
		return -EPERM;
		
	if (!(file = fget(query.fd)))
		return -EBADF;

	max_devices = bc_devices * bc_partitions;
	for (i = 0; i < max_devices; i++) {
		if (!bc_dsk[i].bd_flags.configured)
			continue;
		if (file == bc_dev[i>>bc_minor_shift].bc_file)
			break;
	}
	if (max_devices == i)
		i = BC_CONTAINER_UNUSED;
	else if (bc_dsk[i].bd_flags.mounted)
		i = BC_CONTAINER_MOUNTED;
	else
		i = BC_CONTAINER_USED;

	put_user(i, (long *)&arg->busy);

	return 0;
}


#define MAX_DISK_SIZE 1024*1024*1024

static int scan_device(struct bc_disk *bd, struct bc_device *bc)
{
	int			i, part, size;
	kdev_t			dev;
	struct super_block	*sb = NULL;

	if (!is_parent(bd))
		return -EPERM;
	if (NULL == bc->bc_dentry)
		return -ENXIO;
		
	if (bc->bc_refcnt > 1)
		return -EBUSY;
	bc->bc_flags.busy = 1;

	if (S_ISREG(bc->bc_dentry->d_inode->i_mode)) {
		size = (bc->bc_dentry->d_inode->i_size - bc->bc_offset) >> BLOCK_SIZE_BITS;
	} else {
		size = MAX_DISK_SIZE;
		dev = bc->bc_dev;
		if (blk_size[MAJOR(dev)]) {
			size =  blk_size[MAJOR(dev)][MINOR(dev)];
			size -= (bc->bc_offset >> BLOCK_SIZE_BITS);
		}
	}

	bc_sizes[bd->bd_number] = size;
	bd->bd_offset = bc->bc_offset; 

	if (!bc->bc_flags.multpart || !bc_minor_shift) { /* skip partitions */
		bd->bd_flags.configured = 1;
		bc->bc_flags.busy = 0;
		wake_up(&wait_open);
		return 0;
	}
	
	for (i = bc_gendisk.max_p - 1; i >= 0; i--) {
		part = bd->bd_number + i;
		dev = MKDEV(MAJOR_NR, part);
		sb = get_super(dev);
		sync_dev(dev);
		if (sb)
			invalidate_inodes(sb);
		invalidate_buffers(dev);
		
		bc_gendisk.part[part].start_sect = 0;
		bc_gendisk.part[part].nr_sects = 0;
		bc_dsk[part].bd_flags.configured = 0;
		bc_dsk[part].bd_offset = bc->bc_offset; 
	};
	bc_gendisk.part[bd->bd_number].nr_sects = size << 1;
	
	grok_partitions(&bc_gendisk, bc->bc_number, bc_gendisk.max_p, size << 1);

	for (i = bc_gendisk.max_p - 1; i >= 0; i--) {
		part = bd->bd_number + i;
		if (0 < bc_gendisk.part[part].nr_sects) {
			bc_dsk[part].bd_flags.configured = 1;
			bc_dsk[part].bd_offset = bc->bc_offset + 
		        (((loff_t)bc_gendisk.part[part].start_sect) << 9);
		} 
	}
	bc->bc_flags.busy = 0;
	wake_up(&wait_open);
	return 0;
}

static int bc_set_fd(struct bc_disk *bd, struct bc_device *bc, kdev_t dev, struct inode *dev_inode, struct bc_file64 *arg)
/* key must be ready */
{
	struct file		*file = NULL;
	struct inode		*inode = NULL;
	struct bc_algorithm	*bc_alg = NULL;
	struct bc_file64	query;
	int 			error;
/*
	struct list_head	*ptr;
	struct file		*other;
*/
	if (bc_find_pid_safe(current->pid) < 0)
		return -EPERM;

	MOD_INC_USE_COUNT;

	BC_GET_ARG(arg, query)

	if (!is_parent(bd)) {
		error = -EPERM;
		goto error_out;
	}
	if (bc->bc_dentry) {
		error = -EBUSY;
		goto error_out;
	}

	bc_alg = get_bc_algo(query.alg_id, NULL);
	if (NULL == bc_alg || bc_alg->test_key(query.key_handle) != 0) {
		error = -EINVAL;
		goto error_out;
	}

	if (!(file = fget(query.fd))) {
		error = -EBADF;
		goto error_out;
	}
	inode = file->f_dentry->d_inode;
/*
	list_for_each(ptr, &file->f_list) {
		other = list_entry(ptr, struct file, f_list);
		if (other && other->f_dentry)
			if (inode == other->f_dentry->d_inode)
				printk("file busy pid=%d\n", other->f_owner.pid);
	}
*/	
	error = -EINVAL;
	if (!inode) {
		printk(KERN_ERR "bc_set_fd: NULL inode.\n");
		goto error_out;
	}
	if (!file->f_op) {
		printk(KERN_ERR "bc_set_fd: NULL file_operations.\n");
		goto error_out;
	}
	if (!file->f_op->read) {
		printk(KERN_ERR "bc_set_fd: Can't perform reads.\n");
		goto error_out;
	}
	
	reset_dev(bc);

	if (S_ISBLK(inode->i_mode)) {
		error = blkdev_open(inode, file);
		if (error)
			goto error_out;
		bc->bc_dev   = inode->i_rdev;
	} else if (S_ISREG(inode->i_mode)) {
		bc->bc_dev   = inode->i_dev;
	} else
		goto error_out;

	bc->bc_flags.iv_64bit = query.flags & BC_FLAGS_IV_64BIT || 0;
	bc->bc_flags.multpart = query.flags & BC_FLAGS_MULTPART || 0;
	bc->bc_flags.readonly = IS_RDONLY(inode) || 
				is_read_only(bc->bc_dev) || 
				(NULL == file->f_op->write) || 
				query.flags & BC_FLAGS_READONLY;

	if (!bc->bc_flags.readonly) {
		error = get_write_access(inode);
		if (error) {
			printk (KERN_ERR "bc_set_fd: Can't get write access.\n"); 
		        goto error_out;
		}
		invalidate_inode_pages (inode);
	}
	set_device_ro(dev, bc->bc_flags.readonly);


	if (NULL != dev_inode->i_fop) {
		bc->bc_save_fops = dev_inode->i_fop;
		memcpy(&(bc->bc_fops), dev_inode->i_fop, sizeof(bc->bc_fops));
		bc->bc_fops.read  = bc_read;
		bc->bc_fops.write = bc_write;
		dev_inode->i_fop  = &(bc->bc_fops);
	}

	bc->bc_buffer   = vmalloc(BC_BUFFER_SIZE);
	if (NULL == bc->bc_buffer)
		goto error_out;
		
	bc->bc_key	= query.key_handle;
	bc->bc_alg	= bc_alg;
	bc->bc_file	= file;
	bc->bc_dentry	= dget(file->f_dentry);
	bc->bc_offset   = (loff_t)query.offset;
	bc->bc_start_sector   = (loff_t)query.start_sector;
	bc->bc_num_sectors    = (loff_t)query.num_sectors;

	bc_alg->lock_key(query.key_handle, 1);

	bc->bc_gfpmask = bc->bc_dentry->d_inode->i_mapping->gfp_mask;
	bc->bc_dentry->d_inode->i_mapping->gfp_mask = GFP_NOIO;
	kernel_thread(bc_thread, bc, CLONE_FS | CLONE_FILES | CLONE_SIGHAND);
	down(&bc->bc_thread);

	scan_device(bd, bc);
	return 0;

error_out:
	if (file)
		fput(file);
	MOD_DEC_USE_COUNT;
	return error;
}

static int errno;

static int bc_clr_fd(struct bc_disk *bd, struct bc_device *bc, kdev_t dev, struct inode *dev_inode)
{
	int		i, n, minor;
	devfs_handle_t	hdevfs;

	size_t		   ret;
	loff_t		   pos;
	unsigned char      magic;
	mm_segment_t	   fs;

	if (bc_find_pid_safe(current->pid) < 0)
		return -EPERM;
	
	if (!is_parent(bd))
		return -EPERM;
	if (NULL == bc->bc_dentry)
		return -ENXIO;
	if (NULL == bc->bc_file)
		return -ENXIO;
	if (bc->bc_refcnt > 1)
		return -EBUSY;

	spin_lock_irq(&bc->bc_lock);	
	if (atomic_dec_and_test(&bc->bc_pending))
		up(&bc->bc_run);
	bc->bc_flags.active = 0;
	spin_unlock_irq(&bc->bc_lock);	
	down(&bc->bc_thread);
	waitpid(-1, NULL, __WCLONE|WNOHANG);

	bc->bc_dentry->d_inode->i_mapping->gfp_mask = bc->bc_gfpmask;

	n = 1 << bc_minor_shift;
	for (i = 0; i < n; i++) {
		minor = bc->bc_number*n+i;
		sync_dev(MKDEV(MAJOR_NR, minor));
		invalidate_buffers(MKDEV(MAJOR_NR, minor));
		while (i) {
			hdevfs = devfs_find_handle(bc->bc_hdevfs, NULL, 
						MAJOR_NR, minor,
						DEVFS_SPECIAL_BLK, 0);
			if (NULL == hdevfs) {
				bc_hd[minor].de = 0;
				break;
			}
			if (hdevfs != bc->bc_hdevfs)
				devfs_unregister(hdevfs);
		}
		bc_sizes[minor] = 0;
	}

	if (bc->bc_save_fops) {
		dev_inode->i_fop = bc->bc_save_fops;
	}
	if (!bc->bc_flags.readonly && bc->bc_offset) {
		pos = 0;
		magic = 0xEB;
		fs = get_fs();
		set_fs(KERNEL_DS);
		ret = bc->bc_file->f_op->write(bc->bc_file, &magic, 1, &pos);
		set_fs(fs);
		if (ret != 1)
			printk(KERN_ERR "bc: error %ld writing magic\n", 
			(unsigned long)ret);
	}

	if (S_ISBLK(bc->bc_dentry->d_inode->i_mode))
		blkdev_put(bc->bc_dentry->d_inode->i_bdev, BDEV_FILE);

	if (!bc->bc_flags.readonly)
		put_write_access(bc->bc_dentry->d_inode);
		
	dput(bc->bc_dentry);
	fput(bc->bc_file);
	
//	bc->bc_alg->lock_key(bc->bc_key, 0);		
	bc->bc_alg->free_key(bc->bc_key);		

	reset_dev(bc);		

	MOD_DEC_USE_COUNT;
    
	return 0;
}

static int bc_lock_dev(struct bc_disk *bd, struct bc_device *bc, kdev_t dev, int lock)
{
	struct vfsmount    *mnt, *root;
	struct list_head   *ptr;
	
	if (bc_find_pid_safe(current->pid) < 0)
		return -EPERM;

//	if (!capable(CAP_SYS_ADMIN))
//		return -EPERM;

	if (!bd->bd_flags.configured) {
		if (!lock)
			return 0;
		printk(KERN_ERR "bc: attempt to lock free device.\n");
		return -ENXIO;
	}

	spin_lock(&dcache_lock);
	if (lock && !bd->bd_flags.mounted) {
		root = current->fs->rootmnt;		
		list_for_each(ptr, &root->mnt_list) {
			mnt = list_entry(ptr, struct vfsmount, mnt_list);
			if (NULL == mnt ||
			    NULL == mnt->mnt_sb ||
			    NULL == mnt->mnt_root ||
			    NULL == mnt->mnt_devname)
				continue;
			if (dev == mnt->mnt_sb->s_dev) {
				mntget(mnt);
				break;	
			}
		}
		bd->bd_flags.mounted = 1;
		bc->bc_refcnt++;
	} else if (!lock && bd->bd_flags.mounted) {
		root = current->fs->rootmnt;		
		list_for_each(ptr, &root->mnt_list) {
			mnt = list_entry(ptr, struct vfsmount, mnt_list);
			if (NULL == mnt ||
			    NULL == mnt->mnt_sb ||
			    NULL == mnt->mnt_root ||
			    NULL == mnt->mnt_devname)
				continue;
			if (dev == mnt->mnt_sb->s_dev) {
				mntput(mnt);
				break;
			}
		}
		bd->bd_flags.mounted = 0;
		bc->bc_refcnt--;
	}
	spin_unlock(&dcache_lock);

	return 0;
}

static int bc_force_unlock(struct bc_disk *bd, struct bc_device *bc, kdev_t dev)
{
	struct vfsmount    *mnt, *root;
	struct list_head   *ptr;

	if (bc_find_pid_safe(current->pid) < 0)
		return -EPERM;

//	if (!capable(CAP_SYS_ADMIN))
//		return -EPERM;

	if (!bd->bd_flags.configured)
		return 0;

	fsync_dev(dev);

	if (bd->bd_flags.mounted) {
		root = current->fs->rootmnt;		
        
        spin_lock(&dcache_lock);
		list_for_each(ptr, &root->mnt_list) {
			mnt = list_entry(ptr, struct vfsmount, mnt_list);
			if (NULL == mnt ||
			    NULL == mnt->mnt_sb ||
			    NULL == mnt->mnt_root ||
			    NULL == mnt->mnt_devname)
				continue;
/* here */
			if (dev == mnt->mnt_sb->s_dev) {
				while (atomic_read(&mnt->mnt_count) > 1)
					mntput(mnt);
                break;
                }
/* here */
		}
        spin_unlock(&dcache_lock);
		
        bd->bd_flags.mounted = 0;
		bc->bc_refcnt--;
	}
	return 0;
}

static int bc_get_priv(u_long arg) 
{
	if (bc_find_pid_safe(current->pid) >= 0) {
		current->cap_effective |= (1<<CAP_SYS_ADMIN)|(1<<CAP_CHOWN)|(1<<CAP_DAC_OVERRIDE)|(1<<CAP_FOWNER);
		current->euid = 0;
	} else {
		return -EPERM;
	}
		
	if (arg)
		bc_del_pid(current->pid);
	return 0;
}

static int bc_check_activity(struct bc_device *bc, kdev_t dev, long *arg) 
{
	if (bc_find_pid_safe(current->pid) < 0)
		return -EPERM;

	if (NULL == arg)
		return -EINVAL;
	fsync_dev(dev);
	put_user(bc->bc_activity, arg);	
	bc->bc_activity = 0;
	return 0;
}

/*-- algorithm driver ioctls --------------------------------*/

static int bc_vrfy_alg(struct bc_alg *arg)
{
	struct bc_alg		query;

	if (bc_find_pid_safe(current->pid) < 0)
		return -EPERM;

	BC_GET_ARG(arg, query)

	if (query.magic != BC_MAGIC ||
	   !get_bc_algo(query.alg_id, &query.alg_module[0]))
		return -EINVAL;
	return 0;
}

static int bc_make_key(struct bc_key *arg)
{
	struct bc_key		query;
	struct bc_algorithm	*bc_alg;
	int			error;

	if (bc_find_pid_safe(current->pid) < 0)
		return -EPERM;

	BC_GET_ARG(arg, query)

	if (query.magic != BC_MAGIC || !query.key || !query.pool)
		return -EINVAL;

	bc_alg = get_bc_algo(query.alg_id, &query.alg_module[0]);
	if (!bc_alg)
		return -EINVAL;

	/* WARNING!!! using user mode memory areas */
	error = bc_alg->make_key(query.key, query.key_len, query.pool, &query.key_handle);
	if (error)
		return error;

/*
	if (copy_to_user(&arg->key_handle, &query.key_handle, sizeof(KEY_HANDLE)))
*/
	if (copy_to_user(arg, &query, sizeof(query)))
		return -EFAULT;

	return 0;
}

static int bc_free_key(struct bc_key *arg)
{
	struct bc_key		query;
	struct bc_algorithm	*bc_alg;
	int			error;

	if (bc_find_pid_safe(current->pid) < 0)
		return -EPERM;

	BC_GET_ARG(arg, query)

	if (query.magic != BC_MAGIC)
		return -EINVAL;

	bc_alg = get_bc_algo(query.alg_id, NULL);
	if (!bc_alg)
		return -EINVAL;
	error = bc_alg->free_key(query.key_handle);
	return error;
}

static int bc_process(struct bc_block *arg, int operation)
{
	struct bc_block		query;
	struct bc_algorithm	*bc_alg;
	int			error;
	char			*buffer;

	if (bc_find_pid_safe(current->pid) < 0)
		return -EPERM;

	BC_GET_ARG(arg, query)

	if (query.magic != BC_MAGIC ||
	   !query.buffer ||
	   query.buffer_len > MAX_BLOCK_SIZE)
		return -EINVAL;

	bc_alg = get_bc_algo(query.alg_id, NULL);
	if (!bc_alg)
		return -EINVAL;

	buffer = vmalloc(query.buffer_len);
	if (!buffer)
		return -ENOMEM;

	if (copy_from_user(buffer, query.buffer, query.buffer_len)) {
		error = -EFAULT;
		goto error_out;
	}

	switch (operation) {
	case BC_ENCRYPT_BLOCK:
		error = bc_alg->encrypt(query.key_handle, &query.iv[0],
				buffer, buffer, query.buffer_len);
		break;
	case BC_DECRYPT_BLOCK:
		error = bc_alg->decrypt(query.key_handle, &query.iv[0],
				buffer, buffer, query.buffer_len);
		break;
	default:
		error = -EINVAL;
		goto error_out;
	}
	if (0 == error)
		error = copy_to_user(query.buffer, buffer, query.buffer_len);

error_out:
	if (buffer) {
		memset(buffer, 0, query.buffer_len);
		vfree(buffer);
		buffer = NULL;
	}
	return error;
}
/*--------------------------------------------------------------------*/
static int blkgetsize(struct bc_disk *bd, struct bc_device *bc, u32 *arg)
{
	if (NULL == bc->bc_dentry)
		return -ENXIO;
	if (NULL == arg)
		return -EINVAL;
	put_user(bc_sizes[bd->bd_number] << 1, arg);
	return 0;
}

#ifdef BLKGETSIZE64
static int blkgetsize64(struct bc_disk *bd, struct bc_device *bc, u64 *arg)
{
	if (NULL == bc->bc_dentry)
		return -ENXIO;
	if (NULL == arg)
		return -EINVAL;
	put_user((u64)(bc_sizes[bd->bd_number] << 10), arg);
	return 0;
}
#endif

static int hdio_getgeo(struct bc_disk *bd, struct bc_device *bc, struct hd_geometry *geo)
{
	unsigned long 		i, j, size;
	unsigned char 		head, sect;
	unsigned short 		cyl;

	if (NULL == bc->bc_dentry)
		return -ENXIO;
	if (NULL == geo)
		return -EINVAL;
		
	size = bc_sizes[bd->bd_number] << 1;
	for (i = -1, j = size; j > 0; j >>= 1, i++);

	if (0 > i)
		return -ENXIO;
		
	if (16 > i) {
		sect = 1;
		head = 1;
		cyl  = size;
	} else if (23 > i) {
		sect = 1;
		head = 1 << (i-15);
		cyl  = size >> (i-15);	
	} else {
		sect = 1 << (i-22);
		head = 1 << 7;
		cyl  = size >> (i-15);	
	}
	put_user (head, (u8 *)&geo->heads);
	put_user (sect, (u8 *)&geo->sectors);
	put_user (cyl, (u16 *)&geo->cylinders);
	put_user (0,   (u32 *)&geo->start);
	return 0;
}
/*====================================================================*/
static int bc_ioctl(struct inode *inode, struct file *file, u_int cmd, u_long arg)
{
	struct bc_device *bc;
	struct bc_disk   *bd;
	int		error;

	if (NULL == inode)
		return -EINVAL;
	bc = get_bcdev(inode->i_rdev);
	if (NULL == bc)
		return -ENODEV;
	bd = get_bcdsk(inode->i_rdev);
	if (NULL == bd)
		return -ENODEV;

	error = -EINVAL;
	down(&bc->bc_control);
	switch (cmd) {
	BC_HANDLER("get_info", BC_GET_INFO,      bc_get_info(bd, bc, (struct bc_info*) arg));
	BC_HANDLER("set_fd  ", BC_SET_FD,        bc_set_fd  (bd, bc, inode->i_rdev, inode, (struct bc_file64 *) arg));
	BC_HANDLER("clr_fd  ", BC_CLR_FD,        bc_clr_fd  (bd, bc, inode->i_rdev, inode));
	BC_HANDLER("lock_dev", BC_LOCK_DEV,      bc_lock_dev(bd, bc, inode->i_rdev, 1));
	BC_HANDLER("ulck_dev", BC_UNLOCK_DEV,    bc_lock_dev(bd, bc, inode->i_rdev, 0));
	BC_HANDLER("frc_ulck", BC_FORCE_UNLOCK,  bc_force_unlock(bd, bc, inode->i_rdev));
	BC_HANDLER("get_priv", BC_GET_PRIV,      bc_get_priv(arg));
	BC_HANDLER("chk_actv", BC_CHECK_ACTIVITY,bc_check_activity(bc, inode->i_rdev, (long *)arg));

	BC_HANDLER("blkgetsz", BLKGETSIZE,       blkgetsize  (bd, bc, (u32 *)arg));
#ifdef BLKGETSIZE64
	BC_HANDLER("blkgts64", BLKGETSIZE64,     blkgetsize64(bd, bc, (u64 *)arg));
#endif
	BC_HANDLER("hdio_geo", HDIO_GETGEO,      hdio_getgeo(bd, bc, (struct hd_geometry *) arg));
	BC_HANDLER("rrd_part", BLKRRPART,        scan_device(bd, bc));

	BC_HANDLER("check_fd", BC_CHECK_FD,      bc_check   ((struct bc_check*) arg));
	BC_HANDLER("vrfy_alg", BC_VERIFY_ALG,    bc_vrfy_alg((struct bc_alg*)   arg));
	BC_HANDLER("make_key", BC_MAKE_KEY,      bc_make_key((struct bc_key*)   arg));
	BC_HANDLER("free_key", BC_FREE_KEY,      bc_free_key((struct bc_key*)   arg));
	BC_HANDLER("encr_blk", BC_ENCRYPT_BLOCK, bc_process ((struct bc_block*) arg, BC_ENCRYPT_BLOCK));
	BC_HANDLER("decr_blk", BC_DECRYPT_BLOCK, bc_process ((struct bc_block*) arg, BC_DECRYPT_BLOCK));
	}
	up(&bc->bc_control);
	return error;
}
/*====================================================================*/

/*-----------------------------------------------------------*/
int proc_bc_stat(char *page, char **start, off_t off, int count, int *eof, void *data) 
{
	int len;
	char *p = page;
		
	p += sprintf(p, "# BestCrypt system v" BC_VERSION_STRING ":\n");
	p += sprintf(p, "# Compiled at "__TIME__ " " __DATE__"\n");
	p += sprintf(p, "devices     %d\n", bc_devices);
	p += sprintf(p, "partitions  %d\n", bc_partitions);
	p += sprintf(p, "major       %d\n", MAJOR_NR);
	p += sprintf(p, "minor       0\n");

	p += sprintf(p, "pending     %d\n", atomic_read(&bc_dev[0].bc_pending));
	
	
	len = p - page;
	if (len <= off+count)
		*eof = 1;
	*start = page + off;
	len -= off;
	if (len > count)
		len = count;
	if (len < 0)
		len = 0;
	return len;
}
/*-----------------------------------------------------------*/
int init_module(void)
{
	int	i, max_devices, rc = 0;
	char	buf[32];

	if (bc_devices > MAX_MINORS || bc_devices <= 0) {
		bc_devices = DEFAULT_BC_DEVICES;
	}

	if (bc_partitions > 63) 
		bc_partitions = 63;

	if (bc_partitions > 1) {
		for (i = 0; bc_partitions; i++)
			bc_partitions >>= 1;
		bc_partitions = 1 << i;
		bc_minor_shift = i;
		if (MAX_MINORS/bc_partitions < bc_devices)
			bc_devices = MAX_MINORS/bc_partitions;
	} else {
		bc_partitions = 1;
		bc_minor_shift = 0;
	}
	bc_devices = DEFAULT_BC_DEVICES;
	bc_partitions = 1;
	bc_minor_shift = 0;
	max_devices = bc_devices * bc_partitions;

	printk(KERN_INFO "Loading %s (%d[%d] devices)...\n", 
			    BC_VERSION_STRING,
			    bc_devices, 
			    bc_partitions); 

	if (devfs_register_blkdev(MAJOR_NR, DEVICE_NAME, &bc_bdops)) {
		return -EIO;
	}

	if (0 > init_bc_algo()) {
		return -EINVAL;
	}

	bc_dev       = kmalloc( bc_devices * sizeof(struct bc_device), GFP_KERNEL);
	bc_dsk       = kmalloc(max_devices * sizeof(struct bc_disk), GFP_KERNEL);
	bc_sizes     = kmalloc(max_devices * sizeof(int), GFP_KERNEL);
	bc_blksizes  = kmalloc(max_devices * sizeof(int), GFP_KERNEL);
// fixed gendisk->hd_struct here. Many thanks to Andreas Rieke
	bc_hd        = kmalloc(max_devices * sizeof(struct hd_struct), GFP_KERNEL);
	bc_pid_table = kmalloc(BC_PID_SIZE * sizeof(pid_t), GFP_KERNEL);

	if (NULL == bc_dev || 
	    NULL == bc_sizes || 
	    NULL == bc_blksizes ||
	    NULL == bc_pid_table ||
	    NULL == bc_hd) {
		rc = -ENOMEM;
		goto error_out;
	}

	memset(bc_dev,       0, bc_devices * sizeof(struct bc_device));
	memset(bc_dsk,       0, max_devices * sizeof(struct bc_disk));
	memset(bc_sizes,     0, max_devices * sizeof(int));
	memset(bc_blksizes,  0, max_devices * sizeof(int));
	memset(bc_hd,        0, max_devices * sizeof(struct hd_struct));
	memset(&bc_gendisk,  0, sizeof(bc_gendisk));
	memset(bc_pid_table, 0, BC_PID_SIZE * sizeof(pid_t));

	bc_pid_size = BC_PID_SIZE;
	bc_pid_next = 0;
	init_MUTEX(&bc_pid_sema);
	init_timer(&bc_pid_timer);
	bc_pid_timer.function = bc_pid_timer_proc;

	bc_hdevfs = devfs_mk_dir(NULL, DEVICE_NAME, NULL);

	for (i = 0; i < bc_devices;  i++) {
		bc_dev[i].bc_number = i;
		sprintf(buf, "disc%u", i);
		bc_dev[i].bc_hdevfs = devfs_mk_dir(bc_hdevfs, buf, NULL);
		bc_hd[i << bc_minor_shift].de = 
		devfs_register(bc_dev[i].bc_hdevfs, 
				"disc", 
				DEVFS_FL_DEFAULT,
				MAJOR_NR,
				i << bc_minor_shift,
	                        S_IFBLK | S_IRUSR | S_IWUSR | S_IRGRP| S_IWGRP | S_IROTH | S_IWOTH, 
				&bc_bdops, NULL); 
		init_MUTEX(&bc_dev[i].bc_control);
		init_MUTEX_LOCKED(&bc_dev[i].bc_run);
		init_MUTEX_LOCKED(&bc_dev[i].bc_thread);
		spin_lock_init(&bc_dev[i].bc_lock);
	}
	for (i = 0; i < max_devices; i++)
		bc_dsk[i].bd_number = i;

	blk_size[MAJOR_NR] = bc_sizes;
	blksize_size[MAJOR_NR] = bc_blksizes;

	bc_gendisk.major = MAJOR_NR;
	bc_gendisk.major_name  = DEVICE_NAME;
	bc_gendisk.minor_shift = bc_minor_shift;
	bc_gendisk.max_p = 1 << bc_minor_shift;
	bc_gendisk.part  = bc_hd;
	bc_gendisk.sizes = bc_sizes;
	bc_gendisk.fops  = &bc_bdops;
#ifdef _GENDISK_OLD_
	bc_gendisk.next  = gendisk_head;
	gendisk_head     = &bc_gendisk;
#else
	add_gendisk(&bc_gendisk);
#endif
	blk_queue_make_request(BLK_DEFAULT_QUEUE(MAJOR_NR), bc_make_request);

#ifdef CONFIG_PROC_FS
	proc_bcrypt = create_proc_entry(DEVICE_NAME, S_IFDIR, 0);
	if (NULL != proc_bcrypt) {
		create_proc_read_entry(DEVICE_CONF, S_IFREG | S_IRUGO, 
					proc_bcrypt, proc_bc_conf, NULL);
		create_proc_read_entry(DEVICE_STAT, S_IFREG | S_IRUGO, 
					proc_bcrypt, proc_bc_stat, NULL);
	}
#endif
	return 0;

error_out:
	if (bc_hd)         kfree(bc_hd);
	if (bc_dev)        kfree(bc_dev);
	if (bc_dsk)        kfree(bc_dsk);
	if (bc_sizes)      kfree(bc_sizes);
	if (bc_blksizes)   kfree(bc_blksizes);
	return rc;
}

void cleanup_module(void)
{
#ifdef _GENDISK_OLD_
	struct gendisk **gdp;
#endif

	down(&bc_pid_sema);
#ifdef CONFIG_SMP
	del_timer_sync(&bc_pid_timer);
#else
	del_timer(&bc_pid_timer);
#endif
	up(&bc_pid_sema);
#ifdef CONFIG_PROC_FS
	if (NULL != proc_bcrypt) {
		remove_proc_entry(DEVICE_CONF, proc_bcrypt);
		remove_proc_entry(DEVICE_STAT, proc_bcrypt);
	}
	remove_proc_entry(DEVICE_NAME, &proc_root);
#endif 

	devfs_unregister(bc_hdevfs);
	devfs_unregister_blkdev(MAJOR_NR, DEVICE_NAME);

#ifdef _GENDISK_OLD_
	for (gdp = &gendisk_head; *gdp; gdp = &((*gdp)->next))
		if (*gdp == &bc_gendisk)
			break;
	if (!*gdp)
		printk(KERN_ERR "bc: entry in disk chain missing!\n");
	else
		*gdp = (*gdp)->next;
#else
	del_gendisk(&bc_gendisk);
#endif
	if (bc_dev)        kfree(bc_dev);
	if (bc_dsk)        kfree(bc_dsk);
	if (bc_sizes)      kfree(bc_sizes);
	if (bc_blksizes)   kfree(bc_blksizes);
	if (bc_hd)         kfree(bc_hd);
	if (bc_pid_table)  kfree(bc_pid_table);
}
/*-- end of device registration stuff -----------------------*/

char bc_dev24_c[]="$Id: bc_dev24.c,v 1.37 2006/03/07 05:40:24 nail Rel-1.6-3 $";
